"""Command line interfaces."""
import json
import logging
import os
import sys

import numpy as np
import toml
import yaml

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None


from .binary_input import (
    InlineForecastInputData,
    JsonOutputDataFromFile,
    OfflineInputData,
    PgdInputData,
    PrepInputData,
    SodaInputData,
)
from .cache import Cache
from .cmd_parsing import (
    parse_args_bufr2json,
    parse_args_create_forcing,
    parse_args_create_namelist,
    parse_args_first_guess_for_oi,
    parse_args_gridpp,
    parse_args_hm2pysurfex,
    parse_args_lsm_file_assim,
    parse_args_masterodb,
    parse_args_merge_qc_data,
    parse_args_modify_forcing,
    parse_args_obs2json,
    parse_args_oi2soda,
    parse_args_plot_points,
    parse_args_qc2obsmon,
    parse_args_set_geo_from_obs_set,
    parse_args_set_geo_from_stationlist,
    parse_args_shape2ign,
    parse_args_surfex_binary,
    parse_args_titan,
    parse_cryoclim_pseudoobs,
    parse_sentinel_obs,
    parse_set_domain,
)
from .configuration import (
    ConfigurationFromHarmonieAndConfigFile,
    ConfigurationFromTomlFile,
)
from .datetime_utils import as_datetime, as_datetime_args, as_timedelta
from .file import PGDFile, PREPFile, SurfexFileVariable, SURFFile
from .forcing import modify_forcing, run_time_loop, set_forcing_config
from .geo import LonLatVal, get_geo_object, set_domain, shape2ign
from .grib import Grib1Variable, Grib2Variable
from .input_methods import create_obsset_file, get_datasources, set_geo_from_obs_set
from .interpolation import horizontal_oi
from .namelist import BaseNamelist, Namelist
from .netcdf import (
    create_netcdf_first_guess_template,
    oi2soda,
    read_cryoclim_nc,
    read_first_guess_netcdf_file,
    read_sentinel_nc,
    write_analysis_netcdf_file,
)
from .obs import Observation, sm_obs_sentinel, snow_pseudo_obs_cryoclim
from .obsmon import write_obsmon_sqlite_file
from .platform import SystemFilePathsFromFile
from .read import ConvertedInput, Converter
from .run import BatchJob, Masterodb, PerturbedOffline, SURFEXBinary
from .titan import (
    TitanDataSet,
    dataset_from_file,
    define_quality_control,
    merge_json_qc_data_sets,
)


def get_geo_and_config_from_cmd(**kwargs):
    """Get geo and config from cmd."""
    if "harmonie" in kwargs and kwargs["harmonie"]:
        config_exp = None
        if "config_exp" in kwargs:
            if kwargs["config_exp"] is not None:
                config_exp = kwargs["config_exp"]
        if config_exp is None:
            config_exp = (
                f"{os.path.abspath(os.path.dirname(__file__))}/cfg/config_exp_surfex.toml"
            )
        logging.info("Using default config from: %s", config_exp)
        config = ConfigurationFromHarmonieAndConfigFile(os.environ, config_exp)
        geo = config.geo
    else:
        if "domain" in kwargs:
            domain = kwargs["domain"]
            if os.path.exists(domain):
                with open(domain, mode="r", encoding="utf-8") as fhandler:
                    geo = get_geo_object(json.load(fhandler))
            else:
                raise FileNotFoundError(domain)
        else:
            geo = None

        if "config" in kwargs:
            config = kwargs["config"]
            if os.path.exists(config):
                config = ConfigurationFromTomlFile(config)
            else:
                raise FileNotFoundError("File not found: " + config)
        else:
            config = None
    return config, geo


def run_first_guess_for_oi(**kwargs):
    """Run first guess for oi."""
    config, geo = get_geo_and_config_from_cmd(**kwargs)

    config_file = kwargs["input_config"]
    if not os.path.exists(config_file):
        raise FileNotFoundError(config_file)

    if "output" in kwargs:
        output = kwargs["output"]
    else:
        raise RuntimeError("No output file provided")

    dtg = kwargs["dtg"]
    validtime = as_datetime(dtg)
    variables = kwargs["variables"]
    variables = variables + ["altitude", "land_area_fraction"]

    cache = Cache(3600)
    f_g = None
    for var in variables:

        inputfile = kwargs.get("inputfile")
        fileformat = kwargs.get("inputformat")
        logging.debug("inputfile: %s", inputfile)
        logging.debug("fileformat: %s", fileformat)

        converter = "none"
        if var == "air_temperature_2m":
            if "t2m_file" in kwargs and kwargs["t2m_file"] is not None:
                inputfile = kwargs["t2m_file"]
            if "t2m_format" in kwargs and kwargs["t2m_format"] is not None:
                fileformat = kwargs["t2m_format"]
            if "t2m_converter" in kwargs and kwargs["t2m_converter"] is not None:
                converter = kwargs["t2m_converter"]
        elif var == "relative_humidity_2m":
            if "rh2m_file" in kwargs and kwargs["rh2m_file"] is not None:
                inputfile = kwargs["rh2m_file"]
            if "rh2m_format" in kwargs and kwargs["rh2m_format"] is not None:
                fileformat = kwargs["rh2m_format"]
            if "rh2m_converter" in kwargs and kwargs["rh2m_converter"] is not None:
                converter = kwargs["rh2m_converter"]
        elif var == "surface_snow_thickness":
            if "sd_file" in kwargs and kwargs["sd_file"] is not None:
                inputfile = kwargs["sd_file"]
            if "sd_format" in kwargs and kwargs["sd_format"] is not None:
                fileformat = kwargs["sd_format"]
            if "sd_converter" in kwargs and kwargs["sd_converter"] is not None:
                converter = kwargs["sd_converter"]
        elif var == "cloud_base":
            if "cb_file" in kwargs and kwargs["cb_file"] is not None:
                inputfile = kwargs["cb_file"]
            if "cb_format" in kwargs and kwargs["cb_format"] is not None:
                fileformat = kwargs["cb_format"]
            if "cb_converter" in kwargs and kwargs["cb_converter"] is not None:
                converter = kwargs["cb_converter"]
        elif var == "surface_soil_moisture":
            if "sm_file" in kwargs and kwargs["sm_file"] is not None:
                inputfile = kwargs["sm_file"]
            if "sm_format" in kwargs and kwargs["sm_format"] is not None:
                fileformat = kwargs["sm_format"]
            if "sm_converter" in kwargs and kwargs["sm_converter"] is not None:
                converter = kwargs["sm_converter"]
        elif var == "altitude":
            if "altitude_file" in kwargs and kwargs["altitude_file"] is not None:
                inputfile = kwargs["altitude_file"]
            if "altitude_format" in kwargs and kwargs["altitude_format"] is not None:
                fileformat = kwargs["altitude_format"]
            if (
                "altitude_converter" in kwargs
                and kwargs["altitude_converter"] is not None
            ):
                converter = kwargs["altitude_converter"]
        elif var == "land_area_fraction":
            if "laf_file" in kwargs and kwargs["laf_file"] is not None:
                inputfile = kwargs["laf_file"]
            if "laf_format" in kwargs and kwargs["laf_format"] is not None:
                fileformat = kwargs["laf_format"]
            if "laf_converter" in kwargs and kwargs["laf_converter"] is not None:
                converter = kwargs["laf_converter"]
        else:
            raise NotImplementedError("Variable not implemented " + var)

        if inputfile is None:
            raise RuntimeError("You must set input file")

        if fileformat is None:
            raise RuntimeError("You must set file format")

        logging.debug(inputfile)
        logging.debug(fileformat)
        logging.debug(converter)
        config = yaml.safe_load(open(config_file, "r", encoding="utf-8"))
        defs = config[fileformat]
        defs.update({"filepattern": inputfile})

        logging.debug("Variable: %s", var)
        logging.debug("Fileformat: %s", fileformat)
        converter_conf = config[var][fileformat]["converter"]
        if converter not in config[var][fileformat]["converter"]:
            logging.debug("config_file: %s", config_file)
            logging.debug("config: %s", config)
            raise RuntimeError(f"No converter {converter} definition found in {config}!")

        initial_basetime = validtime - as_timedelta(seconds=10800)
        converter = Converter(
            converter, initial_basetime, defs, converter_conf, fileformat
        )
        field = ConvertedInput(geo, var, converter).read_time_step(validtime, cache)
        field = np.reshape(field, [geo.nlons, geo.nlats])

        # Create file
        if f_g is None:
            n_x = geo.nlons
            n_y = geo.nlats
            f_g = create_netcdf_first_guess_template(variables, n_x, n_y, output, geo=geo)
            epoch = float(
                (validtime - as_datetime_args(year=1970, month=1, day=1)).total_seconds()
            )
            f_g.variables["time"][:] = epoch
            f_g.variables["longitude"][:] = np.transpose(geo.lons)
            f_g.variables["latitude"][:] = np.transpose(geo.lats)
            f_g.variables["x"][:] = list(range(0, n_x))
            f_g.variables["y"][:] = list(range(0, n_y))

        if var == "altitude":
            field[field < 0] = 0

        if np.isnan(np.sum(field)):
            fill_nan_value = f_g.variables[var].getncattr("_FillValue")
            logging.info("Field %s got Nan. Fill with: %s", var, str(fill_nan_value))
            field[np.where(np.isnan(field))] = fill_nan_value

        f_g.variables[var][:] = np.transpose(field)

    if f_g is not None:
        f_g.close()


def run_masterodb(**kwargs):
    """Run masterodb."""
    logging.debug("ARGS: %s", kwargs)
    config, geo = get_geo_and_config_from_cmd(**kwargs)

    if "config" in kwargs:
        del kwargs["config"]

    system_file_paths = kwargs["system_file_paths"]
    if os.path.exists(system_file_paths):
        system_file_paths = SystemFilePathsFromFile(system_file_paths)
    else:
        raise FileNotFoundError("File not found: " + system_file_paths)
    del kwargs["system_file_paths"]

    binary = kwargs["binary"]
    rte = kwargs["rte"]
    wrapper = kwargs["wrapper"]
    namelist_path = kwargs["namelist_path"]
    force = kwargs["force"]
    mode = kwargs["mode"]
    output = kwargs["output"]
    archive = kwargs["archive"]
    only_archive = kwargs["only_archive"]
    print_namelist = kwargs["print_namelist"]

    check_existence = True
    if "tolerate_missing" in kwargs:
        if kwargs["tolerate_missing"]:
            check_existence = False

    dtg = None
    if "dtg" in kwargs:
        if kwargs["dtg"] is not None and isinstance(kwargs["dtg"], str):
            dtg = as_datetime(kwargs["dtg"])
            kwargs.update({"dtg": dtg})

    # TODO
    perturbed_file_pattern = None
    if perturbed_file_pattern in kwargs:
        perturbed_file_pattern = kwargs["perturbed_file_pattern"]

    pgd_file_path = kwargs["pgd"]
    prep_file_path = kwargs["prep"]

    if os.path.exists(rte):
        with open(rte, mode="r", encoding="utf-8") as file_handler:
            rte = json.load(file_handler)
        my_batch = BatchJob(rte, wrapper=wrapper)
    else:
        raise FileNotFoundError

    my_archive = None
    if archive is not None:
        if os.path.exists(archive):
            my_archive = JsonOutputDataFromFile(archive)
        else:
            raise FileNotFoundError

    if mode == "forecast":
        input_data = InlineForecastInputData(
            config, system_file_paths, check_existence=check_existence
        )
        mode = "offline"
    elif mode == "canari":
        input_data = SodaInputData(
            config,
            system_file_paths,
            check_existence=check_existence,
            perturbed_file_pattern=perturbed_file_pattern,
            dtg=dtg,
        )
        mode = "soda"
    else:
        raise NotImplementedError(mode + " is not implemented!")

    blocks = False
    if blocks:
        my_settings = Namelist(
            mode, config, namelist_path, dtg=dtg, fcint=3
        ).get_namelist()
    else:
        my_settings = BaseNamelist(
            mode, config, namelist_path, dtg=dtg, fcint=3
        ).get_namelist()
    geo.update_namelist(my_settings)

    # Create input
    my_format = my_settings["nam_io_offline"]["csurf_filetype"]
    my_pgdfile = my_settings["nam_io_offline"]["cpgdfile"]
    my_prepfile = my_settings["nam_io_offline"]["cprepfile"]
    my_surffile = my_settings["nam_io_offline"]["csurffile"]
    lfagmap = False
    if "lfagmap" in my_settings["nam_io_offline"]:
        lfagmap = my_settings["nam_io_offline"]["lfagmap"]

    logging.debug("%s %s", my_pgdfile, lfagmap)
    # Not run binary
    masterodb = None
    if not only_archive:
        # Normal dry or wet run
        exists = False
        if output is not None:
            exists = os.path.exists(output)
        if not exists or force:

            if binary is None:
                my_batch = None

            my_pgdfile = PGDFile(
                my_format,
                my_pgdfile,
                input_file=pgd_file_path,
                lfagmap=lfagmap,
                masterodb=True,
            )
            my_prepfile = PREPFile(
                my_format,
                my_prepfile,
                input_file=prep_file_path,
                lfagmap=lfagmap,
                masterodb=True,
            )
            surffile = SURFFile(
                my_format,
                my_surffile,
                archive_file=output,
                lfagmap=lfagmap,
                masterodb=True,
            )

            masterodb = Masterodb(
                my_pgdfile,
                my_prepfile,
                surffile,
                my_settings,
                input_data,
                binary=binary,
                print_namelist=print_namelist,
                batch=my_batch,
                archive_data=my_archive,
            )

        else:
            logging.info("%s already exists!", output)

    if archive is not None:
        if masterodb is not None:
            masterodb.archive_output()
        else:
            logging.info("Masterodb is None")


def run_surfex_binary(mode, **kwargs):
    """Run a surfex binary."""
    logging.debug("ARGS: %s", kwargs)
    config, geo = get_geo_and_config_from_cmd(**kwargs)

    system_file_paths = kwargs["system_file_paths"]
    if os.path.exists(system_file_paths):
        system_file_paths = SystemFilePathsFromFile(system_file_paths)
    else:
        raise FileNotFoundError("File not found: " + system_file_paths)

    if "forcing_dir" in kwargs:
        system_file_paths.add_system_file_path("forcing_dir", kwargs["forcing_dir"])

    pgd = False
    prep = False
    perturbed = False
    need_pgd = True
    need_prep = True
    check_existence = True
    if "tolerate_missing" in kwargs:
        if kwargs["tolerate_missing"]:
            check_existence = False

    prep_input_file = None
    if "prep_file" in kwargs:
        prep_input_file = kwargs["prep_file"]
    prep_input_pgdfile = None
    if "prep_pgdfile" in kwargs:
        prep_input_pgdfile = kwargs["prep_pgdfile"]
    prep_input_filetype = None
    if "prep_filetype" in kwargs:
        prep_input_filetype = kwargs["prep_filetype"]
    prep_input_pgdfiletype = None
    if "prep_pgdfiletype" in kwargs:
        prep_input_pgdfiletype = kwargs["prep_pgdfiletype"]

    # TODO
    perturbed_file_pattern = None
    if perturbed_file_pattern in kwargs:
        perturbed_file_pattern = kwargs["perturbed_file_pattern"]

    dtg = None
    if "dtg" in kwargs:
        if kwargs["dtg"] is not None and isinstance(kwargs["dtg"], str):
            dtg = as_datetime(kwargs["dtg"])
            kwargs.update({"dtg": dtg})

    logging.debug("kwargs: %s", str(kwargs))
    if mode == "pgd":
        pgd = True
        need_pgd = False
        need_prep = False
        input_data = PgdInputData(
            config, system_file_paths, check_existence=check_existence
        )
    elif mode == "prep":
        prep = True
        need_prep = False
        input_data = PrepInputData(
            config,
            system_file_paths,
            check_existence=check_existence,
            prep_file=prep_input_file,
            prep_pgdfile=prep_input_pgdfile,
        )
    elif mode == "offline":
        input_data = OfflineInputData(
            config, system_file_paths, check_existence=check_existence
        )
    elif mode == "soda":
        input_data = SodaInputData(
            config,
            system_file_paths,
            check_existence=check_existence,
            masterodb=kwargs["masterodb"],
            perturbed_file_pattern=perturbed_file_pattern,
            dtg=dtg,
        )
    elif mode == "perturbed":
        perturbed = True
        input_data = OfflineInputData(
            config, system_file_paths, check_existence=check_existence
        )
    else:
        raise NotImplementedError(mode + " is not implemented!")

    binary = kwargs["binary"]
    rte = kwargs["rte"]
    wrapper = kwargs["wrapper"]
    namelist_path = kwargs["namelist_path"]
    force = kwargs["force"]
    output = kwargs["output"]
    archive = kwargs["archive"]
    print_namelist = kwargs["print_namelist"]
    masterodb = kwargs["masterodb"]
    logging.debug("masterodb %s", masterodb)
    forc_zs = False
    if "forc_zs" in kwargs:
        forc_zs = kwargs["forc_zs"]

    pgd_file_path = None
    if need_pgd:
        pgd_file_path = kwargs["pgd"]

    prep_file_path = None
    if need_prep:
        prep_file_path = kwargs["prep"]

    pert = None
    if "pert" in kwargs:
        pert = kwargs["pert"]
    negpert = False
    if "negpert" in kwargs:
        negpert = kwargs["negpert"]

    if os.path.exists(rte):
        with open(rte, mode="r", encoding="utf-8") as file_handler:
            rte = json.load(file_handler)
        my_batch = BatchJob(rte, wrapper=wrapper)
    else:
        raise FileNotFoundError("File not found: " + rte)

    my_archive = None
    if archive is not None:
        if os.path.exists(archive):
            my_archive = JsonOutputDataFromFile(archive)
        else:
            raise FileNotFoundError("File not found: " + archive)

    if not os.path.exists(output) or force:
        blocks = False
        if blocks:
            my_settings = Namelist(
                mode,
                config,
                namelist_path,
                forc_zs=forc_zs,
                prep_file=prep_input_file,
                prep_filetype=prep_input_filetype,
                prep_pgdfile=prep_input_pgdfile,
                prep_pgdfiletype=prep_input_pgdfiletype,
                dtg=dtg,
                fcint=3,
            ).get_namelist()
        else:
            my_settings = BaseNamelist(
                mode,
                config,
                namelist_path,
                forc_zs=forc_zs,
                prep_file=prep_input_file,
                prep_filetype=prep_input_filetype,
                prep_pgdfile=prep_input_pgdfile,
                prep_pgdfiletype=prep_input_pgdfiletype,
                dtg=dtg,
                fcint=3,
            ).get_namelist()
        geo.update_namelist(my_settings)

        # Create input
        my_format = my_settings["nam_io_offline"]["csurf_filetype"]
        my_pgdfile = my_settings["nam_io_offline"]["cpgdfile"]
        my_prepfile = my_settings["nam_io_offline"]["cprepfile"]
        my_surffile = my_settings["nam_io_offline"]["csurffile"]
        lfagmap = False
        if "lfagmap" in my_settings["nam_io_offline"]:
            lfagmap = my_settings["nam_io_offline"]["lfagmap"]

        logging.debug("pgdfile=%s lfagmap=%s %s", my_pgdfile, lfagmap, pgd_file_path)
        if need_pgd:
            logging.debug("Need pgd")
            my_pgdfile = PGDFile(
                my_format,
                my_pgdfile,
                input_file=pgd_file_path,
                lfagmap=lfagmap,
                masterodb=masterodb,
            )

        if need_prep:
            logging.debug("Need prep")
            my_prepfile = PREPFile(
                my_format,
                my_prepfile,
                input_file=prep_file_path,
                lfagmap=lfagmap,
                masterodb=masterodb,
            )

        surffile = None
        if need_prep and need_pgd:
            logging.debug("Need pgd and prep")
            surffile = SURFFile(
                my_format,
                my_surffile,
                archive_file=output,
                lfagmap=lfagmap,
                masterodb=masterodb,
            )

        if perturbed:
            PerturbedOffline(
                binary,
                my_batch,
                my_prepfile,
                pert,
                my_settings,
                input_data,
                pgdfile=my_pgdfile,
                surfout=surffile,
                archive_data=my_archive,
                print_namelist=print_namelist,
                negpert=negpert,
            )
        elif pgd:
            my_pgdfile = PGDFile(
                my_format,
                my_pgdfile,
                input_file=pgd_file_path,
                archive_file=output,
                lfagmap=lfagmap,
                masterodb=masterodb,
            )
            SURFEXBinary(
                binary,
                my_batch,
                my_pgdfile,
                my_settings,
                input_data,
                archive_data=my_archive,
                print_namelist=print_namelist,
            )
        elif prep:
            my_prepfile = PREPFile(
                my_format,
                my_prepfile,
                archive_file=output,
                lfagmap=lfagmap,
                masterodb=masterodb,
            )
            SURFEXBinary(
                binary,
                my_batch,
                my_prepfile,
                my_settings,
                input_data,
                pgdfile=my_pgdfile,
                archive_data=my_archive,
                print_namelist=print_namelist,
            )
        else:
            SURFEXBinary(
                binary,
                my_batch,
                my_prepfile,
                my_settings,
                input_data,
                pgdfile=my_pgdfile,
                surfout=surffile,
                archive_data=my_archive,
                print_namelist=print_namelist,
            )

    else:
        logging.info("%s already exists!", output)


def run_create_namelist(**kwargs):
    """Create a namelist."""
    logging.debug("ARGS: %s", kwargs)
    config, geo = get_geo_and_config_from_cmd(**kwargs)
    mode = kwargs.get("mode")

    system_file_paths = kwargs["system_file_paths"]
    if os.path.exists(system_file_paths):
        system_file_paths = SystemFilePathsFromFile(system_file_paths)
    else:
        raise FileNotFoundError("File not found: " + system_file_paths)

    if "forcing_dir" in kwargs:
        system_file_paths.add_system_file_path("forcing_dir", kwargs["forcing_dir"])

    prep_input_file = None
    if "prep_file" in kwargs:
        prep_input_file = kwargs["prep_file"]
    prep_input_pgdfile = None
    if "prep_pgdfile" in kwargs:
        prep_input_pgdfile = kwargs["prep_pgdfile"]
    prep_input_filetype = None
    if "prep_filetype" in kwargs:
        prep_input_filetype = kwargs["prep_filetype"]
    prep_input_pgdfiletype = None
    if "prep_pgdfiletype" in kwargs:
        prep_input_pgdfiletype = kwargs["prep_pgdfiletype"]

    # TODO
    perturbed_file_pattern = None
    if perturbed_file_pattern in kwargs:
        perturbed_file_pattern = kwargs["perturbed_file_pattern"]

    output = kwargs.get("output")
    if output is None:
        output = "OPTIONS.nam"
    dtg = None
    if "dtg" in kwargs:
        if kwargs["dtg"] is not None and isinstance(kwargs["dtg"], str):
            dtg = as_datetime(kwargs["dtg"])
            kwargs.update({"dtg": dtg})

    logging.debug("kwargs: %s", str(kwargs))

    namelist_path = kwargs["namelist_path"]
    forc_zs = False
    if "forc_zs" in kwargs:
        forc_zs = kwargs["forc_zs"]

    if kwargs.get("method") == "blocks":
        my_settings = Namelist(
            mode,
            config,
            namelist_path,
            forc_zs=forc_zs,
            prep_file=prep_input_file,
            geo=geo,
            prep_filetype=prep_input_filetype,
            prep_pgdfile=prep_input_pgdfile,
            prep_pgdfiletype=prep_input_pgdfiletype,
            dtg=dtg,
            fcint=3,
        ).get_namelist()
    else:
        my_settings = BaseNamelist(
            mode,
            config,
            namelist_path,
            forc_zs=forc_zs,
            prep_file=prep_input_file,
            prep_filetype=prep_input_filetype,
            prep_pgdfile=prep_input_pgdfile,
            prep_pgdfiletype=prep_input_pgdfiletype,
            geo=geo,
            dtg=dtg,
            fcint=3,
        ).get_namelist()
    geo.update_namelist(my_settings)
    if os.path.exists(output):
        os.remove(output)
    my_settings.write(output)


def run_gridpp(**kwargs):
    """Gridpp."""
    var = kwargs["var"]
    input_file = kwargs["input_file"]
    output_file = None
    if "output_file" in kwargs:
        output_file = kwargs["output_file"]
    hlength = kwargs["hlength"]
    vlength = 100000
    if "vlength" in kwargs:
        vlength = kwargs["vlength"]
    wlength = 0
    if "wlength" in kwargs:
        wlength = kwargs["wlength"]
    max_locations = 20
    if "max_locations" in kwargs:
        max_locations = kwargs["max_locations"]
    elev_gradient = 0
    if "elev_gradient" in kwargs:
        elev_gradient = kwargs["elev_gradient"]
    if elev_gradient != -0.0065 and elev_gradient != 0:
        raise RuntimeError("Not a valid elevation gradient")
    epsilon = 0.25
    if "epsilon" in kwargs:
        epsilon = kwargs["epsilon"]
    minvalue = None
    if "minvalue" in kwargs:
        minvalue = kwargs["minvalue"]
    maxvalue = None
    if "maxvalue" in kwargs:
        maxvalue = kwargs["maxvalue"]
    only_diff = False
    if "only_diff" in kwargs:
        only_diff = kwargs["only_diff"]
    obs_file = kwargs["obs_file"]

    # Get input fields
    geo, validtime, background, glafs, gelevs = read_first_guess_netcdf_file(
        input_file, var
    )

    an_time = validtime
    # Read OK observations
    observations = dataset_from_file(an_time, obs_file, qc_flag=0)
    logging.info("Found %s observations with QC flag == 0", str(len(observations.lons)))

    field = horizontal_oi(
        geo,
        background,
        observations,
        gelevs,
        hlength=hlength,
        vlength=vlength,
        wlength=wlength,
        structure_function="Barnes",
        max_locations=max_locations,
        elev_gradient=elev_gradient,
        epsilon=epsilon,
        minvalue=minvalue,
        maxvalue=maxvalue,
        interpol="bilinear",
        only_diff=only_diff,
    )

    if output_file is not None:
        write_analysis_netcdf_file(
            output_file, field, var, validtime, gelevs, glafs, new_file=True, geo=geo
        )


def run_titan(**kwargs):
    """Titan."""
    __, domain_geo = get_geo_and_config_from_cmd(**kwargs)
    if "domain_geo" in kwargs:
        if domain_geo is not None:
            logging.info("Override domain with domain_geo")
        with open(kwargs["domain_geo"], mode="r", encoding="utf-8") as fhandler:
            domain_geo = json.load(fhandler)

    blacklist = None
    if "blacklist" in kwargs:
        blacklist = kwargs["blacklist"]
    elif "blacklist_file" in kwargs:
        if kwargs["blacklist_file"] is not None:
            blacklist = json.load(open(kwargs["blacklist_file"], "r", encoding="utf-8"))

    if "input_file" in kwargs:
        input_file = kwargs["input_file"]
        if os.path.exists(input_file):
            settings = json.load(open(input_file, "r", encoding="utf-8"))
        else:
            raise FileNotFoundError("Could not find input file " + input_file)
    else:
        if "input_data" in kwargs:
            settings = kwargs["input_data"]
        else:
            raise RuntimeError("You must specify input_file or input_data")

    tests = kwargs["tests"]
    output_file = None
    if "output_file" in kwargs:
        output_file = kwargs["output_file"]
    indent = None
    if "indent" in kwargs:
        indent = kwargs["indent"]

    an_time = kwargs["dtg"]
    if isinstance(an_time, str):
        an_time = as_datetime(an_time)
    var = kwargs["variable"]

    tests = define_quality_control(
        tests, settings[var], an_time, domain_geo=domain_geo, blacklist=blacklist
    )
    logging.debug("Settings: %s", settings)
    datasources = get_datasources(an_time, settings[var]["sets"])
    data_set = TitanDataSet(var, settings[var], tests, datasources, an_time)
    data_set.perform_tests()

    if output_file is not None:
        data_set.write_output(output_file, indent=indent)


def run_oi2soda(**kwargs):
    """Oi2soda."""
    t2m_file = kwargs["t2m_file"]
    rh2m_file = kwargs["rh2m_file"]
    sd_file = kwargs["sd_file"]
    sm_file = kwargs["sm_file"]
    output = kwargs["output"]

    t2m = None
    if t2m_file is not None:
        t2m = {"file": t2m_file, "var": kwargs["t2m_var"]}
    rh2m = None
    if rh2m_file is not None:
        rh2m = {"file": rh2m_file, "var": kwargs["rh2m_var"]}
    s_d = None
    if sd_file is not None:
        s_d = {"file": sd_file, "var": kwargs["sd_var"]}
    s_m = None
    if sm_file is not None:
        s_m = {"file": sm_file, "var": kwargs["sm_var"]}

    dtg = as_datetime(kwargs["dtg"])
    oi2soda(dtg, t2m=t2m, rh2m=rh2m, s_d=s_d, s_m=s_m, output=output)


def run_hm2pysurfex(**kwargs):
    """Harmonie to pysurfex."""
    pysurfex_config = kwargs["config"]

    output = None
    if "output" in kwargs:
        output = kwargs["output"]

    environment = os.environ
    if "environment" in kwargs:
        environment_file = kwargs["environment"]
        environment.update(json.load(open(environment_file, "r", encoding="utf-8")))

    # Create configuration
    config = ConfigurationFromHarmonieAndConfigFile(environment, pysurfex_config)

    if output is None:
        logging.info("Config settings %s", config.settings)
    else:
        with open(output, "w", encoding="utf-8") as fhandler:
            toml.dump(config.settings, fhandler)


def run_plot_points(**kwargs):
    """Point plots."""
    geo_file = None
    if "geo" in kwargs:
        geo_file = kwargs["geo"]
    validtime = None
    if kwargs["validtime"] is not None:
        validtime = as_datetime(kwargs["validtime"])
    variable = None
    if "variable" in kwargs:
        variable = kwargs["variable"]
    filepattern = None
    if "inputfile" in kwargs:
        filepattern = kwargs["inputfile"]
    inputtype = kwargs["inputtype"]
    output = kwargs["output"]
    interpolator = "nearest"
    if "interpolator" in kwargs:
        interpolator = kwargs["interpolator"]

    geo = None
    if geo_file is not None:
        domain_json = json.load(open(geo_file, "r", encoding="utf-8"))
        geo = get_geo_object(domain_json)

    contour = True
    if "no_contour" in kwargs:
        no_contour = kwargs["no_contour"]
        if no_contour:
            contour = False

    var = "field_to_read"
    if inputtype == "grib1":

        if filepattern is None:
            raise RuntimeError("You must provide a filepattern")

        par = kwargs["indicatorOfParameter"]
        ltp = kwargs["levelType"]
        lev = kwargs["level"]
        tri = kwargs["timeRangeIndicator"]

        gribvar = Grib1Variable(par, ltp, lev, tri)
        title = (
            "grib1:" + gribvar.generate_grib_id() + " " + validtime.strftime("%Y%m%d%H")
        )
        var_dict = {
            "filepattern": filepattern,
            "fcint": 10800,
            "file_inc": 10800,
            "offset": 0,
            "parameter": par,
            "type": ltp,
            "level": lev,
            "tri": tri,
            "interpolator": interpolator,
        }

    elif inputtype == "grib2":

        if filepattern is None:
            raise RuntimeError("You must provide a filepattern")

        discipline = kwargs["discipline"]
        parameter_category = kwargs["parameterCategory"]
        parameter_number = kwargs["parameterNumber"]
        level_type = kwargs["levelType"]
        level = kwargs["level"]
        type_of_statistical_processing = kwargs["typeOfStatisticalProcessing"]

        gribvar = Grib2Variable(
            discipline,
            parameter_category,
            parameter_number,
            level_type,
            level,
            tsp=type_of_statistical_processing,
        )
        logging.debug(inputtype)
        logging.debug(gribvar)
        logging.debug(validtime)
        title = (
            f"{inputtype}: {gribvar.generate_grib_id()} {validtime.strftime('%Y%m%d%H')}"
        )

        var_dict = {
            "fcint": 10800,
            "file_inc": 10800,
            "offset": 0,
            "filepattern": filepattern,
            "discipline": discipline,
            "parameterCategory": parameter_category,
            "parameterNumber": parameter_number,
            "levelType": level_type,
            "level": level,
            "typeOfStatisticalProcessing": type_of_statistical_processing,
        }

    elif inputtype == "netcdf":

        if variable is None:
            raise RuntimeError("You must provide a variable")
        if filepattern is None:
            raise RuntimeError("You must provide a filepattern")

        title = "netcdf: " + variable + " " + validtime.strftime("%Y%m%d%H")
        var_dict = {
            "name": variable,
            "filepattern": filepattern,
            "fcint": 10800,
            "file_inc": 10800,
            "offset": 0,
            "interpolator": interpolator,
        }

    elif inputtype == "surfex":

        if variable is None:
            raise RuntimeError("You must provide a variable")
        if filepattern is None:
            raise RuntimeError("You must provide a filepattern")

        basetime = kwargs["sfx_basetime"]
        patches = kwargs["sfx_patches"]
        layers = kwargs["sfx_layers"]
        datatype = kwargs["sfx_datatype"]
        interval = kwargs["sfx_interval"]
        geo_sfx_input = kwargs["sfx_geo_input"]
        geo_input = None
        if geo_sfx_input is not None:
            domain_json = json.load(open(geo_sfx_input, "r", encoding="utf-8"))
            geo_input = get_geo_object(domain_json)

        sfx_var = SurfexFileVariable(
            variable,
            validtime=validtime,
            patches=patches,
            layers=layers,
            basetime=basetime,
            interval=interval,
            datatype=datatype,
        )

        title = inputtype + ": " + sfx_var.print_var()
        var_dict = {
            "varname": variable,
            "filepattern": filepattern,
            "patches": patches,
            "layers": layers,
            "datatype": datatype,
            "interval": interval,
            "basetime": basetime,
            "geo_input": geo_input,
            "fcint": 10800,
            "file_inc": 10800,
            "offset": 0,
            "interpolator": interpolator,
        }

    elif inputtype == "obs":

        contour = False
        if variable is None:
            raise RuntimeError("You must provide a variable")

        obs_input_type = kwargs["obs_type"]
        if obs_input_type is None:
            raise RuntimeError("You must provide an obs type")

        if geo is None:
            obs_time = as_datetime(kwargs["validtime"])
            varname = variable
            inputfile = kwargs["inputfile"]
            geo = set_geo_from_obs_set(
                obs_time, obs_input_type, varname, inputfile, lonrange=None, latrange=None
            )

        var_dict = {
            "filetype": obs_input_type,
            "varname": variable,
            "filepattern": filepattern,
            "filenames": [filepattern],
            "fcint": 10800,
            "file_inc": 10800,
            "offset": 0,
        }
        title = inputtype + ": var=" + variable + " type=" + obs_input_type

    else:
        raise NotImplementedError

    defs = {var: {inputtype: {"converter": {"none": var_dict}}}}
    converter_conf = defs[var][inputtype]["converter"]

    if geo is None:
        raise RuntimeError("No geo is set")

    cache = Cache(-1)
    converter = "none"
    converter = Converter(converter, validtime, defs, converter_conf, inputtype)
    field = ConvertedInput(geo, var, converter).read_time_step(validtime, cache)

    if field is None:
        raise RuntimeError("No field read")

    logging.debug(
        "npoints=%s nlons=%s nlats=%s contour=%s field.shape=%s",
        geo.npoints,
        geo.nlons,
        geo.nlats,
        contour,
        field.shape,
    )
    if geo.npoints != geo.nlons and geo.npoints != geo.nlats:
        if contour:
            field = np.reshape(field, [geo.nlons, geo.nlats])
    else:
        contour = False

    if plt is None:
        raise ModuleNotFoundError("Matplotlib is needed to plot")
    logging.debug(
        "lons.shape=%s lats.shape=%s field.shape=%s",
        geo.lons.shape,
        geo.lats.shape,
        field.shape,
    )
    if contour:
        plt.contourf(geo.lons, geo.lats, field)
    else:
        plt.scatter(geo.lonlist, geo.latlist, c=field)

    plt.title(title)
    plt.colorbar()
    if output is None:
        plt.show()
    else:
        logging.info("Saving figure in %s", output)
        plt.savefig(output)


def set_geo_from_stationlist(**kwargs):
    """Set geometry from station list."""
    stationlist = kwargs["stationlist"]
    lonrange = None
    if "lonrange" in kwargs:
        lonrange = kwargs["lonrange"]
    latrange = None
    if "latrange" in kwargs:
        latrange = kwargs["latrange"]
    if lonrange is None:
        lonrange = [-180, 180]
    if latrange is None:
        latrange = [-90, 90]

    lons = []
    lats = []
    if os.path.exists(stationlist):
        stids = json.load(open(stationlist, "r", encoding="utf-8"))
    else:
        raise FileNotFoundError("Station list does not exist!")

    for stid in stids:
        lon, lat = Observation.get_pos_from_stid(stationlist, [stid])
        lon = lon[0]
        lat = lat[0]
        if lonrange[0] <= lon <= lonrange[1] and latrange[0] <= lat <= latrange[1]:
            lon = round(lon, 5)
            lat = round(lat, 5)
            lons.append(lon)
            lats.append(lat)

    d_x = ["0.3"] * len(lons)
    geo_json = {
        "nam_pgd_grid": {"cgrid": "LONLATVAL"},
        "nam_lonlatval": {"xx": lons, "xy": lats, "xdx": d_x, "xdy": d_x},
    }
    return LonLatVal(geo_json)


def sentinel_obs(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_sentinel_obs(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ sentinel_obs ******************")
    fg_file = kwargs["fg_file"]
    infiles = kwargs["infiles"]
    step = kwargs["thinning"]
    output = kwargs["output"]
    varname = kwargs["varname"]
    indent = kwargs["indent"]

    grid_lons, grid_lats, grid_sm_class = read_sentinel_nc(infiles)
    fg_geo, validtime, grid_sm_fg, __, __ = read_first_guess_netcdf_file(fg_file, varname)
    q_c = sm_obs_sentinel(
        validtime, grid_sm_class, grid_lons, grid_lats, step, fg_geo, grid_sm_fg
    )
    q_c.write_output(output, indent=indent)


def qc2obsmon(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_qc2obsmon(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ qc2obsmon ******************")
    write_obsmon_sqlite_file(**kwargs)


def prep(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "prep")
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ prep ******************")
    run_surfex_binary("prep", **kwargs)


def plot_points(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_plot_points(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ plot_points ******************")
    run_plot_points(**kwargs)


def pgd(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "pgd")
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ pgd ******************")
    run_surfex_binary("pgd", **kwargs)


def perturbed_offline(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "perturbed")
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ offline ******************")
    run_surfex_binary("perturbed", **kwargs)


def offline(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_surfex_binary(argv, "offline")
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ offline ******************")
    run_surfex_binary("offline", **kwargs)


def cli_oi2soda(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_oi2soda(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ oi2soda ******************")
    run_oi2soda(**kwargs)


def cli_modify_forcing(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_modify_forcing(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ modify_forcing ******************")
    modify_forcing(**kwargs)


def cli_merge_qc_data(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_merge_qc_data(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ merge_qc_data ******************")

    qc_data = merge_json_qc_data_sets(kwargs.get("validtime"), kwargs.get("filenames"))
    qc_data.write_output(kwargs.get("output"), indent=kwargs.get("indent"))


def masterodb(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_masterodb(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ masterodb ******************")
    run_masterodb(**kwargs)


def hm2pysurfex(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_hm2pysurfex(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ hm2pysurfex ******************")
    run_hm2pysurfex(**kwargs)


def gridpp(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_gridpp(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ gridpp ******************")
    run_gridpp(**kwargs)


def dump_environ(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    with open("rte.json", mode="w", encoding="utf-8") as file_handler:
        json.dump(os.environ.copy(), file_handler)


def first_guess_for_oi(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_first_guess_for_oi(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ FirstGuess4gridpp ******************")
    run_first_guess_for_oi(**kwargs)


def cryoclim_pseudoobs(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_cryoclim_pseudoobs(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ cryoclim_pseudoobs ******************")

    fg_file = kwargs["fg_file"]
    infiles = kwargs["infiles"]
    step = kwargs["thinning"]
    output = kwargs["output"]
    varname = kwargs["varname"]
    indent = kwargs["indent"]

    grid_lons, grid_lats, grid_snow_class = read_cryoclim_nc(infiles)
    fg_geo, validtime, grid_snow_fg, __, __ = read_first_guess_netcdf_file(
        fg_file, varname
    )
    q_c = snow_pseudo_obs_cryoclim(
        validtime, grid_snow_class, grid_lons, grid_lats, step, fg_geo, grid_snow_fg
    )
    q_c.write_output(output, indent=indent)


def create_namelist(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_create_namelist(argv)
    debug = kwargs.get("debug")
    mode = kwargs.get("mode")
    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ %s ******************", mode)
    run_create_namelist(**kwargs)


def create_lsm_file_assim(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileNotFoundError: Domain file not found

    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_lsm_file_assim(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ create_lsm_fil ******************")

    domain = kwargs["domain"]
    logging.debug("domain=%s", domain)
    if os.path.exists(domain):
        domain_json = json.load(open(domain, "r", encoding="utf-8"))
        geo = get_geo_object(domain_json)
    else:
        raise FileNotFoundError(domain)
    validtime = as_datetime(kwargs["dtg"])

    # TODO Move to a method outside cli
    cache = Cache(3600)
    inputfile = kwargs["file"]
    fileformat = kwargs["fileformat"]
    converter = kwargs["converter"]
    output = kwargs["output"]

    var = kwargs["var"]

    defs = {
        "filepattern": inputfile,
        "fileformat": fileformat,
        "filetype": "surf",
        "fcint": 10800,
        "offset": 0,
    }

    logging.debug("%s %s", var, fileformat)
    converter_conf = {"none": {"name": var, "varname": var}}

    var = "LSM"
    initial_basetime = validtime - as_timedelta(seconds=10800)
    converter = Converter(converter, initial_basetime, defs, converter_conf, fileformat)
    field = ConvertedInput(geo, var, converter).read_time_step(validtime, cache)
    field = np.reshape(field, [geo.nlons, geo.nlats])
    field = np.transpose(field)

    with open(output, mode="w", encoding="utf-8") as file_handler:
        for lat in range(0, geo.nlats):
            for lon in range(0, geo.nlons):
                file_handler.write(str(field[lat, lon]) + "\n")


def create_forcing(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_create_forcing(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ create_forcing ******************")
    options, var_objs, att_objs = set_forcing_config(**kwargs)
    run_time_loop(options, var_objs, att_objs)


def bufr2json(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_bufr2json(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ bufr2json ******************")
    logging.warning("This method is depreciated. Please use create_obsset_file")
    variables = kwargs.get("vars")
    bufrfile = kwargs.get("bufr")
    output = kwargs.get("output")
    valid_dtg = as_datetime(kwargs.get("dtg"))
    valid_range = as_timedelta(seconds=kwargs.get("valid_range"))
    label = kwargs.get("label")
    indent = kwargs.get("indent")
    lonrange = kwargs.get("lonrange")
    latrange = kwargs.get("latrange")

    create_obsset_file(
        valid_dtg,
        "bufr",
        variables,
        bufrfile,
        output,
        pos_t_range=valid_range,
        lonrange=lonrange,
        latrange=latrange,
        label=label,
        indent=indent,
    )


def obs2json(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_obs2json(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ obs2json ******************")
    obs_type = kwargs.get("obs_type")
    variables = kwargs.get("vars")
    inputfile = kwargs.get("inputfile")
    output = kwargs.get("output")
    obs_time = as_datetime(kwargs.get("obs_time"))
    label = kwargs.get("label")
    unit = kwargs.get("unit")
    level = kwargs.get("level")
    obtypes = kwargs.get("obtypes")
    subtypes = kwargs.get("subtypes")
    pos_t_range = kwargs.get("pos_t_range")
    neg_t_range = kwargs.get("neg_t_range")
    indent = kwargs.get("indent")
    lonrange = kwargs.get("lonrange")
    latrange = kwargs.get("latrange")
    if pos_t_range is not None:
        pos_t_range = as_timedelta(seconds=pos_t_range)
    if neg_t_range is not None:
        neg_t_range = as_timedelta(seconds=neg_t_range)

    create_obsset_file(
        obs_time,
        obs_type,
        variables,
        inputfile,
        output,
        pos_t_range=pos_t_range,
        neg_t_range=neg_t_range,
        lonrange=lonrange,
        latrange=latrange,
        label=label,
        unit=unit,
        level=level,
        indent=indent,
        obtypes=obtypes,
        subtypes=subtypes,
    )


def cli_set_domain(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.

    Raises:
        FileNotFoundError: File not found
        RuntimeError: Domain not provided

    """
    if argv is None:
        argv = sys.argv[1:]

    args = parse_set_domain(argv)
    debug = args.debug

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_domain ******************")
    domain = args.domain
    domains = args.domains
    output = args.output
    indent = args.indent
    harmonie_mode = args.harmonie
    if os.path.exists(domains):
        with open(domains, mode="r", encoding="utf-8") as file_handler:
            domains = json.load(file_handler)
        domain_json = set_domain(domains, domain, hm_mode=harmonie_mode)
        if domain_json is not None:
            with open(output, mode="w", encoding="utf-8") as file_handler:
                json.dump(domain_json, file_handler, indent=indent)
        else:
            raise RuntimeError("Domain not provided")
    else:
        raise FileNotFoundError


def cli_set_geo_from_obs_set(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_set_geo_from_obs_set(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_geo_from_obs_set ******************")

    validtime = as_datetime(kwargs["validtime"])
    geo = set_geo_from_obs_set(
        validtime,
        kwargs["obs_type"],
        kwargs["variable"],
        kwargs["inputfile"],
        kwargs["lonrange"],
        kwargs["latrange"],
    )
    output = kwargs["output"]
    with open(output, mode="w", encoding="utf-8") as file_handler:
        json.dump(geo.json, file_handler)


def cli_set_geo_from_stationlist(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_set_geo_from_stationlist(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ set_geo_from_stationlist ******************")
    geo = set_geo_from_stationlist(**kwargs)
    output = kwargs["output"]
    with open(output, mode="w", encoding="utf-8") as file_handler:
        json.dump(geo.json, file_handler)


def cli_shape2ign(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_shape2ign(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ shape2ign ******************")
    catchment = kwargs.get("catchment")
    infile = kwargs.get("infile")
    output = kwargs.get("output")
    indent = kwargs.get("indent")
    ref_proj = kwargs.get("ref_proj")

    shape2ign(catchment, infile, output, ref_proj, indent=indent)


def soda(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]

    kwargs = parse_args_surfex_binary(argv, "soda")
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ soda ******************")
    run_surfex_binary("soda", **kwargs)


def titan(argv=None):
    """Command line interface.

    Args:
        argv(list, optional): Arguments. Defaults to None.
    """
    if argv is None:
        argv = sys.argv[1:]
    kwargs = parse_args_titan(argv)
    debug = kwargs.get("debug")

    if debug:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(pathname)s:%(lineno)s %(message)s",
            level=logging.DEBUG,
        )
    else:
        logging.basicConfig(
            format="%(asctime)s %(levelname)s %(message)s", level=logging.INFO
        )
    logging.info("************ titan ******************")
    run_titan(**kwargs)



"""Surfex file related stuff."""
import abc
import logging
import os
import re
import shutil

import netCDF4
import numpy as np
import pyproj

from .datetime_utils import as_datetime, as_datetime_args, as_timedelta, isdatetime
from .fa import Fa
from .geo import IGN, ConfProj, LonLatReg, LonLatVal
from .interpolation import Interpolation
from .util import remove_existing_file


class SurfexIO(object):
    """Abstract Surfex IO class.

    Used for internal surfex file format files.

    """

    def __init__(self, filename, geo, extension):
        """Construct a surfex IO object.

        Args:
            filename (str): Name of file
            geo (surfex.Geometry): Geometry
            extension (str): File extension
        """
        self.filename = filename
        self.geo = geo
        self.extension = extension

    @abc.abstractmethod
    def field(self, var, validtime=None):
        """Abstract method to read field.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Raises:
            NotImplementedError: Must be implemented by child class.

        """
        raise NotImplementedError("This method is not implemented for this class!")

    @abc.abstractmethod
    def points(self, var, geo_out, validtime=None, interpolation="bilinear"):
        """Abstract method to read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geometry): Surfex geometry to interpolate to.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type.. Defaults to "nearest".

        Raises:
            NotImplementedError: Must be implemented by child class.

        """
        raise NotImplementedError("This method is not implemented for this class!")

    @staticmethod
    def interpolate_field(field, geo_in, geo_out, interpolation="bilinear"):
        """Interpolate a field to points.

        Args:
            field (np.darray): Field to interpolate
            geo_in (surfex.Geo): Input geometry
            geo_out (surfex.Geo): Target geometry
            interpolation (str, optional): Interpolation type. Defaults to "bilinear".

        Returns:
            tuple: (np.array, surfex.Interpolator)

        """
        interpolator = Interpolation(interpolation, geo_in, geo_out)
        field = interpolator.interpolate(field)
        return field, interpolator


class SurfexSurfIO(object):
    """Surfex surf file.

    Class for surfex restart files.

    """

    def __init__(
        self, surfexfile, csurf_filetype, input_file=None, symlink=True, archive_file=None
    ):
        """Construct the surfex surf file.

        Args:
            surfexfile (str): Internal file name of surfex file.
            csurf_filetype (str): File type
            input_file (str, optional): File name to be used as surfexfile. Defaults to None.
            symlink (bool, optional): Symlink input_file to surfexfile. Defaults to True.
            archive_file (str, optional): Location to store the result. Defaults to None.

        """
        self.filename = surfexfile.filename
        self.csurf_filetype = csurf_filetype
        self.need_pgd = False
        logging.debug("Input file: %s", input_file)
        self.input_file = input_file
        self.symlink = symlink
        if self.input_file is not None:
            if self.symlink:
                self.symlink_input()
            else:
                self.copy_input()
        self.archive_file = archive_file

    def symlink_input(self):
        """Symlink the input file."""
        if self.input_file is not None:
            f_out = os.getcwd() + "/" + self.filename
            logging.debug("input_file: %s file_out: %s", self.input_file, f_out)
            remove_existing_file(self.input_file, f_out)
            if os.path.abspath(self.input_file) != f_out:
                logging.info("Symlink %s -> %s", self.input_file, f_out)
                os.symlink(self.input_file, f_out)

    def copy_input(self):
        """Copy the input file."""
        if self.input_file is not None:
            f_out = os.getcwd() + "/" + self.filename
            remove_existing_file(self.input_file, f_out)
            if os.path.abspath(self.input_file) != f_out:
                logging.info("Copy %s -> %s", self.input_file, f_out)
                shutil.copy2(self.input_file, f_out)

    def archive_output_file(self):
        """Archive the output file."""
        if self.archive_file is not None:
            dirname = os.path.dirname(os.path.abspath(self.archive_file))
            os.makedirs(dirname, exist_ok=True)
            f_in = os.getcwd() + "/" + self.filename
            if os.path.abspath(self.archive_file) != f_in:
                logging.info("Move %s to %s", f_in, self.archive_file)
                if os.path.islink(self.archive_file):
                    os.unlink(self.archive_file)
                if os.path.isfile(self.archive_file):
                    os.remove(self.archive_file)
                if os.path.isdir(self.archive_file):
                    shutil.rmtree(self.archive_file)
                shutil.move(f_in, self.archive_file)


class PGDFile(SurfexSurfIO):
    """PGD file."""

    def __init__(
        self,
        csurf_filetype,
        cpgdfile,
        input_file=None,
        symlink=True,
        archive_file=None,
        lfagmap=False,
        masterodb=False,
    ):
        """Construct PGD file object.

        Args:
            csurf_filetype (str): File type
            cpgdfile (str): Name of the PGD file
            input_file (str, optional): Input file. Defaults to None.
            symlink (bool, optional): Symlink input_file to surfexfile. Defaults to True.
            archive_file (str, optional): Location to store the result. Defaults to None.
            lfagmap (bool, optional): File use LFAGMAP. Defaults to False.
            masterodb (bool, optional): File produced by masterodb. Defaults to False.

        """
        logging.debug("PGDFile")
        logging.debug("%s %s %s", cpgdfile, csurf_filetype, masterodb)

        cpgdfile = get_surfex_io_object(
            cpgdfile,
            filetype="surf",
            fileformat=csurf_filetype,
            lfagmap=lfagmap,
            masterodb=masterodb,
        )

        SurfexSurfIO.__init__(
            self,
            cpgdfile,
            csurf_filetype,
            input_file=input_file,
            archive_file=archive_file,
            symlink=symlink,
        )
        self.need_pgd = False


class PREPFile(SurfexSurfIO):
    """PREP file."""

    def __init__(
        self,
        csurf_filetype,
        cprepfile,
        input_file=None,
        symlink=True,
        archive_file=None,
        lfagmap=False,
        masterodb=False,
    ):
        """Construct PREP file object.

        Args:
            csurf_filetype (str): File type
            cprepfile (str): Name of the PREP file
            input_file (str, optional): Input file. Defaults to None.
            symlink (bool, optional): Symlink input_file to surfexfile. Defaults to True.
            archive_file (str, optional): Location to store the result. Defaults to None.
            lfagmap (bool, optional): File use LFAGMAP. Defaults to False.
            masterodb (bool, optional): File produced by masterodb. Defaults to False.

        """
        logging.debug("PREPFile %s", input_file)
        cprepfile = get_surfex_io_object(
            cprepfile,
            filetype="surf",
            fileformat=csurf_filetype,
            lfagmap=lfagmap,
            masterodb=masterodb,
        )

        SurfexSurfIO.__init__(
            self,
            cprepfile,
            csurf_filetype,
            input_file=input_file,
            archive_file=archive_file,
            symlink=symlink,
        )
        self.need_pgd = True


class SURFFile(SurfexSurfIO):
    """SURFOUT file."""

    def __init__(
        self,
        csurf_filetype,
        csurffile,
        archive_file=None,
        input_file=None,
        lfagmap=False,
        masterodb=False,
    ):
        """Construct SURFOUT file object.

        Result of a surfex binary.

        Args:
            csurf_filetype (str): File type
            csurffile (str): Name of the PREP file
            input_file (str, optional): Input file. Defaults to None.
            archive_file (str, optional): Location to store the result. Defaults to None.
            lfagmap (bool, optional): File use LFAGMAP. Defaults to False.
            masterodb (bool, optional): File produced by masterodb. Defaults to False.

        """
        logging.debug("SURFFile")
        csurffile = get_surfex_io_object(
            csurffile,
            filetype="surf",
            fileformat=csurf_filetype,
            lfagmap=lfagmap,
            masterodb=masterodb,
        )

        SurfexSurfIO.__init__(
            self,
            csurffile,
            csurf_filetype,
            input_file=input_file,
            archive_file=archive_file,
        )
        self.need_pgd = True


class SurfexFileVariable(object):
    """Surfex Variable."""

    def __init__(
        self,
        varname,
        validtime=None,
        patches=1,
        layers=1,
        basetime=None,
        interval=None,
        datatype="float",
        tiletype="FULL",
    ):
        """Construct a surfex file variable.

        Args:
            varname (str): Variable name.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            patches (int, optional): Number of patches. Defaults to 1.
            layers (int, optional): Number of layers. Defaults to 1.
            basetime (datetime.datetime, optional): _description_. Defaults to None.
            interval (int, optional): Interval. Defaults to None.
            datatype (str, optional): Data type for variable. Defaults to "float".
            tiletype (str, optional): Tiletype

        """
        self.varname = varname
        self.type = datatype
        self.patches = patches
        self.layers = layers
        self.basetime = basetime
        self.interval = interval
        self.validtime = validtime
        self.datatype = datatype
        self.tiletype = tiletype

    def print_var(self):
        """Print variable information."""
        return self.varname


def get_surfex_io_object(
    fname, filetype="surf", fileformat=None, geo=None, lfagmap=False, masterodb=False
):
    """Get the surfexIO object.

    Args:
        fname (str): File name.
        filetype (str, optional): File type. Defaults to "surf".
        fileformat (str, optional): File format. Defaults to None.
        geo (surfex.Geo, optional): Geometry. Defaults to None.
        lfagmap (bool, optional): File use LFAGMAP. Defaults to False.
        masterodb (bool, optional): File produced by masterodb. Defaults to False.

    Raises:
        RuntimeError: Invalid filetype
        NotImplementedError: Filetype not implemented
        RuntimeError: Format needs a geometry

    Returns:
        SurfexIO: SurfexIO object.

    """
    logging.debug("get_surfex_io_object")
    if filetype is not None:
        if (
            filetype.lower() != "surf"
            and filetype.lower() != "ts"
            and filetype.lower() != "forcing"
        ):
            raise RuntimeError(
                "Invalid filetype: " + filetype + " Allowed: surf/ts/forcing"
            )

    if fileformat is None:
        fileformat, filetype = guess_file_format(fname, filetype)

    if fileformat.lower() == "ascii":
        if filetype.lower() == "surf":
            obj = AsciiSurfexFile(fname, geo=geo)
        elif filetype.lower() == "forcing":
            raise NotImplementedError("Not implemented yet")
        else:
            raise NotImplementedError

    elif fileformat.lower() == "nc":
        if filetype.lower() == "surf":
            obj = NCSurfexFile(fname, geo=geo)
        else:
            raise NotImplementedError
    elif fileformat.lower() == "netcdf":
        if filetype.lower() == "ts":
            if geo is None:
                raise RuntimeError("Format NetCDF needs a geometry")
            obj = NetCDFSurfexFile(fname, geo)
        elif filetype.lower() == "forcing":
            if geo is None:
                raise RuntimeError(
                    "Format NetCDF needs a geometry for reading forcing files"
                )
            obj = ForcingFileNetCDF(fname, geo)
        else:
            raise NotImplementedError
    elif fileformat.lower() == "texte":
        if geo is None:
            raise RuntimeError("Format TEXTE needs a geometry")
        obj = TexteSurfexFile(fname, geo)
    elif fileformat.lower() == "fa":
        if filetype.lower() == "surf":
            obj = FaSurfexFile(fname, geo=geo, lfagmap=lfagmap, masterodb=masterodb)
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError("Format not implemented: " + fileformat)

    logging.debug("Returning object: %s", obj)
    return obj


def guess_file_format(fname, ftype=None):
    """Guess the file format.

    Args:
        fname (str): Filename
        ftype (str, optional): Filetype if known/wished. Defaults to None.

    Raises:
        RuntimeError: Can not-auto decide filetype for files
        RuntimeError: Filetype and/or format not set

    Returns:
        tuple: fileformat, filetype

    """
    f_n = str(os.path.basename(fname))
    ext = str(f_n.rsplit(".", maxsplit=1)[-1])
    if ftype is None:
        logging.info("Trying to guess the file type")
        needles = ["PREP", "PGD"]
        for needle in needles:
            if re.search(needle, f_n):
                ftype = "surf"

        if ext.endswith("nc"):
            ftype = "surf"
        needles = [".*PROGNOSTIC.*", ".*DIAGNOSTICS.*", "SURF_ATM.*"]
        for needle in needles:
            if re.search(needle, f_n):
                ftype = "ts"
        for __ in needles:
            if ext.endswith("TXT"):
                ftype = "ts"
        needles = ["Forc_.*", "FORCING.*"]
        for needle in needles:
            if re.search(needle, f_n):
                ftype = "forcing"

        if re.search("SURFOUT.*", f_n) and ftype is None:
            raise RuntimeError(
                "Can not-auto decide filetype for files called SURFOUT.*.txt. "
                + "Specify either surf or ts"
            )

    fileformat = None
    logging.info("Trying to guess the file format from extension: %s", ext)
    if ext.endswith("txt"):
        fileformat = "ascii"
    if ext.endswith("TXT"):
        fileformat = "texte"
    if ext.endswith("nc"):
        if ftype == "surf":
            fileformat = "nc"
        else:
            fileformat = "netcdf"
    if ext.endswith("fa"):
        fileformat = "fa"
    if ext.endswith("sfx"):
        fileformat = "fa"

    if ftype is None or fileformat is None:
        raise RuntimeError(
            "Filetype and/or format not set: " + str(ftype) + " & " + str(fileformat)
        )
    logging.info("Filetype: %s format: %s", ftype, fileformat)

    return fileformat, ftype


class AsciiSurfexFile(SurfexIO):
    """Input from an ASCII surfex file (.txt)."""

    def __init__(self, filename, geo=None):
        """Construct the ASCII object.

        Args:
            filename (str): Filename
            geo(surfex.geo.Geo, optional): Geometry, Defaults to None.

        """
        suffix = SurfFileTypeExtension("ASCII").suffix
        self.filename = filename

        if geo is None:
            geo = self.get_geo()

        if not filename.endswith(suffix):
            filename = filename + suffix

        SurfexIO.__init__(self, filename, geo, "txt")

    def get_geo(self):
        """Get the geometry object.

        Raises:
            FileNotFoundError: _description_
            RuntimeError: No grid found
            NotImplementedError: _description_

        Returns:
            surfex.Geometry: Surfex geometry
        """
        if not os.path.isfile(self.filename):
            raise FileNotFoundError("File does not exist: " + str(self.filename))

        grid = self.read("GRID_TYPE", "FULL", "string")
        if grid is None:
            raise RuntimeError("No grid found")

        if grid == "IGN":
            domain = {
                "nam_ign": {
                    "clambert": self.read("LAMBERT", "&FULL", "integer"),
                    "npoints": self.read("NPOINTS", "&FULL", "integer"),
                    "xx": self.read("XX", "&FULL", "float"),
                    "xy": self.read("XY", "&FULL", "float"),
                    "xdx": self.read("XDX", "&FULL", "float"),
                    "xdy": self.read("XY", "&FULL", "float"),
                    "xx_llcorner": self.read("XX_LLCORNER", "&FULL", "float"),
                    "xy_llcorner": self.read("XY_LLCORNER", "&FULL", "float"),
                    "xcellsize": self.read("XCELLSIZE", "&FULL", "float"),
                    "ncols": self.read("NCOLS", "&FULL", "integer"),
                    "nrows": self.read("NROWS", "&FULL", "integer"),
                }
            }
            return IGN(domain)

        elif grid == "LONLATVAL":
            domain = {
                "nam_lonlatval": {
                    "xx": self.read("XX", "&FULL", "float"),
                    "xy": self.read("XY", "&FULL", "float"),
                    "xdx": self.read("DX", "&FULL", "float"),
                    "xdy": self.read("DY", "&FULL", "float"),
                }
            }
            return LonLatVal(domain)

        elif grid == "LONLAT REG":
            domain = {
                "nam_lonlat_reg": {
                    "xlonmin": self.read("LONMIN", "&FULL", "float"),
                    "xlatmin": self.read("LATMIN", "&FULL", "float"),
                    "xlonmax": self.read("LONMAX", "&FULL", "float"),
                    "xlatmax": self.read("LATMAX", "&FULL", "float"),
                    "nlon": self.read("NLON", "&FULL", "integer"),
                    "nlat": self.read("NLAT", "&FULL", "integer"),
                }
            }
            return LonLatReg(domain)

        elif grid == "CONF PROJ":
            lon0 = self.read("LON0", "&FULL", "float")
            lat0 = self.read("LAT0", "&FULL", "float")
            n_x = self.read("IMAX", "&FULL", "integer")
            n_y = self.read("JMAX", "&FULL", "integer")
            d_x = self.read("XX", "&FULL", "float")
            d_y = self.read("XX", "&FULL", "float")
            if d_x.shape[0] > 1:
                d_x = d_x[1] - d_x[0]
            if d_y.shape[0] > 1:
                d_y = d_y[1] - d_y[0]
            ll_lon = self.read("LONORI", "&FULL", "float")
            ll_lat = self.read("LATORI", "&FULL", "float")

            logging.info(
                "lon0=%s lat0=%s n_x=%s, n_y=%s d_x=%s dy=%s, ll_lon=%s, ll_lat=%s",
                lon0,
                lat0,
                n_x,
                n_y,
                d_x,
                d_y,
                ll_lon,
                ll_lat,
            )
            earth = 6.37122e6
            proj_string = (
                f"+proj=lcc +lat_0={str(lat0)} +lon_0={str(lon0)} +lat_1={str(lat0)} "
                f"+lat_2={str(lat0)} +units=m +no_defs +R={str(earth)}"
            )

            proj = pyproj.CRS.from_string(proj_string)
            wgs84 = pyproj.CRS.from_string("EPSG:4326")
            x_0, y_0 = pyproj.Transformer.from_crs(wgs84, proj, always_xy=True).transform(
                ll_lon, ll_lat
            )
            x_c = x_0 + 0.5 * (n_x - 1) * d_x
            y_c = y_0 + 0.5 * (n_y - 1) * d_y
            lonc, latc = pyproj.Transformer.from_crs(
                proj, wgs84, always_xy=True
            ).transform(x_c, y_c)

            domain = {
                "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                "nam_conf_proj_grid": {
                    "xloncen": lonc,
                    "xlatcen": latc,
                    "nimax": n_x,
                    "njmax": n_y,
                    "xdx": d_x,
                    "xdy": d_x,
                    "ilone": 0,
                    "ilate": 0,
                },
            }
            return ConfProj(domain)
        else:
            raise NotImplementedError("Grid " + str(grid[0]) + " not implemented!")

    def read(self, read_par, read_tile, datatype):
        """Read the file.

        Args:
            read_par (str): Parameter to read
            read_tile (str): Tile to read
            datatype (str): Datatype

        Raises:
            NotImplementedError: Datatype not implemented
            RuntimeError: Could not read datatype

        Returns:
            numpy.array: Values read

        """
        # Add & if not given
        if read_tile.find("&") < 0:
            read_tile = "&" + read_tile
        file = open(self.filename, mode="r", encoding="utf-8")
        read_desc = False
        read_value = False
        values = []
        for line in file:
            words = line.split()
            print(read_value, read_desc)
            print(words)
            if len(words) > 0:
                if read_value and not read_desc:
                    if words[0].find("&") < 0:
                        try:
                            if datatype.lower() == "float":
                                for word in words:
                                    val = float(word.replace("D", "E"))
                                    if val == 1e20:
                                        val = np.nan
                                    values.append(val)
                            elif datatype.lower() == "string":
                                str_words = []
                                for word in words:
                                    str_words.append(word)
                                values = " ".join(str_words)
                            elif (
                                datatype.lower() == "integer" or datatype.lower() == "int"
                            ):
                                for word in words:
                                    values.append(int(word))
                            elif (
                                datatype.lower() == "logical"
                                or datatype.lower() == "bool"
                            ):
                                for word in words:
                                    if word.lower().strip()[0] == "t":
                                        values = True
                                    else:
                                        values = False
                            else:
                                raise NotImplementedError(
                                    "Type not implemented " + str(datatype)
                                )
                        except ValueError:
                            raise RuntimeError(
                                f"Conversion from {str(words)} to {str(datatype)} "
                                "does not work! Try a different datatype!"
                            ) from ValueError

                if read_desc:
                    read_desc = False
                    read_value = True

                if words[0].find("&") >= 0:
                    tile = words[0]
                    par = words[1]

                    read_value = False
                    if (
                        tile.strip().lower() == read_tile.lower()
                        and par.lower() == read_par.lower()
                    ):
                        read_desc = True
                        read_value = False
                        logging.info("Found: %s %s", str(tile), str(par))

            # Description could be empty
            else:
                if read_desc:
                    read_desc = False
                    read_value = True

        if isinstance(values, list):
            if len(values) == 0:
                logging.info("No values found for %s", read_par)
                return None

            if len(values) > 1:
                values = np.asarray(values)
            else:
                values = values[0]
        logging.info("Returning values: %s", values)
        return values

    def field(self, var, validtime=None):
        """Read a field in file.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Returns:
            np.darray: Field, surfex.Geo in read file

        """
        read_par = var.varname
        read_tile = var.tiletype
        datatype = var.datatype
        field = self.read(read_par, read_tile, datatype)

        geo_in = self.get_geo()
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest"):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geo): Surfex geometry for points.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type. Defaults to "nearest".

        Returns:
            np.darray: Interpolated points

        """
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


class NCSurfexFile(SurfexIO):
    """NetCDF surfex file (restart type)."""

    def __init__(self, filename, geo=None):
        """Construct NC file object.

        Args:
            filename (str): File name
            geo (surfex.Geo, optional): Surfex geometry. Defaults to None.

        """
        suffix = SurfFileTypeExtension("NC").suffix
        if not filename.endswith(suffix):
            filename = filename + suffix

        self.filename = filename
        if geo is None:
            geo = self.get_geo()

        SurfexIO.__init__(self, filename, geo, "nc")

    def get_geo(self):
        """Get geometry in file.

        Raises:
            NotImplementedError: Grid not implemented

        Returns:
            surfex.Geometry: Surfex geometry in file

        """
        try:
            f_h = netCDF4.Dataset(self.filename, "r")
        except OSError:
            logging.debug("Geo not open %s as a netCDF file", self.filename)
            return None
        cgrid = str(netCDF4.chartostring(f_h["GRID_TYPE"][:])).strip()
        if cgrid == "CONF PROJ":
            lon0 = float(f_h["LON0"][0])
            lat0 = float(f_h["LAT0"][0])
            n_x = int(f_h["IMAX"][0])
            n_y = int(f_h["JMAX"][0])
            d_x = float(f_h["DX"][0][0])
            d_y = float(f_h["DY"][0][0])

            ll_lon = float(f_h["LONORI"][0])
            ll_lat = float(f_h["LATORI"][0])
            earth = 6.37122e6
            proj_string = (
                f"+proj=lcc +lat_0={str(lat0)} +lon_0={str(lon0)} +lat_1={str(lat0)} "
                f"+lat_2={str(lat0)} +units=m +no_defs +R={str(earth)}"
            )

            proj = pyproj.CRS.from_string(proj_string)
            wgs84 = pyproj.CRS.from_string("EPSG:4326")
            x_0, y_0 = pyproj.Transformer.from_crs(wgs84, proj, always_xy=True).transform(
                ll_lon, ll_lat
            )
            x_c = x_0 + 0.5 * (n_x + 1) * d_x
            y_c = y_0 + 0.5 * (n_y + 1) * d_y
            lonc, latc = pyproj.Transformer.from_crs(
                proj, wgs84, always_xy=True
            ).transform(x_c, y_c)

            domain = {
                "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                "nam_conf_proj_grid": {
                    "xloncen": lonc,
                    "xlatcen": latc,
                    "nimax": n_x,
                    "njmax": n_y,
                    "xdx": d_x,
                    "xdy": d_y,
                    "ilone": 0,
                    "ilate": 0,
                },
            }
            return ConfProj(domain)
        elif cgrid == "IGN":
            domain = {
                "nam_ign": {
                    "clambert": f_h["CLAMBERT"][0],
                    "xx": f_h["XX"][:],
                    "xy": f_h["XY"][:],
                    "xdx": f_h["DX"][:],
                    "xdy": f_h["DY"][:],
                }
            }
            return IGN(domain)

        elif cgrid == "LONLATVAL":
            domain = {
                "nam_lonlatval": {
                    "xx": f_h["XX"][:],
                    "xy": f_h["XY"][:],
                    "xdx": f_h["DX"][:],
                    "xdy": f_h["DY"][:],
                }
            }
            return LonLatVal(domain)

        elif cgrid == "LONLAT REG":
            domain = {
                "nam_lonlatval_reg": {
                    "lonmin": f_h["LONMIN"][0],
                    "latmin": f_h["LATMIN"][0],
                    "lonmax": f_h["LONMAX"][0],
                    "latmax": f_h["LATMAX"][0],
                    "nlon": f_h["NLON"][0],
                    "nlat": f_h["NLAT"][0],
                    "reg_lon": f_h["REG_LON"][0],
                    "reg_lat": f_h["REG_LAT"][0],
                }
            }
            return LonLatReg(domain)
        else:
            raise NotImplementedError(cgrid + " is not implemented")

    def field(self, var, validtime=None):
        """Read field.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            validtime (_type_, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: Mismatch in times in file and the wanted time

        Returns:
            np.darray: Field, surfex.Geo in read file

        """
        f_h = netCDF4.Dataset(self.filename, "r")
        if validtime is None:
            pass
        else:
            time_in_file = None
            try:
                year = f_h["DTCUR-YEAR"][0]
                month = f_h["DTCUR-MONTH"][0]
                day = f_h["DTCUR-DAY"][0]
                time = int(f_h["DTCUR-TIME"][0])
                hour = int(time / 3600)

                # TODO minutes
                time_in_file = as_datetime_args(
                    year=year, month=month, day=day, hour=hour
                )
            except IndexError:
                logging.warning("Could not checking time")

            if time_in_file is not None:
                if validtime != time_in_file:
                    logging.error("%s %s", time_in_file, validtime)
                    raise RuntimeError("Mismatch in times in file and the wanted time")

        geo_in = self.get_geo()
        field = f_h[var.varname][:]
        fillvalue = f_h[var.varname].getncattr("_FillValue")
        logging.info("Set %s to nan", fillvalue)
        field = field.filled(np.nan)

        # Reshape to fortran 2D style
        field = np.transpose(field)
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest"):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geo): Surfex geometry for points.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type. Defaults to "nearest".

        Returns:
            np.darray: Interpolated points

        """
        field, geo_in = self.field(var, validtime=validtime)
        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


class FaSurfexFile(SurfexIO):
    """FA surfex file."""

    def __init__(self, filename, geo=None, masterodb=False, lfagmap=True):
        """Construct the surfex FA file.

        Args:
            filename (str): File name.
            geo (surfex.Geo, optional): Surfex geometry. Defaults to None.
            masterodb (bool, optional): _description_. Defaults to False.
            lfagmap (bool, optional): _description_. Defaults to True.

        """
        extension = SurfFileTypeExtension("FA", masterodb=masterodb, lfagmap=lfagmap)
        extension_suffix = extension.suffix

        # Surfex binaries use .fa suffix for all input files no matter if they are lfagmap files

        if not filename.endswith(extension_suffix):
            filename = filename + extension_suffix

        SurfexIO.__init__(self, filename, geo, extension)
        self.lfagmap = lfagmap

    def field(self, var, validtime=None):
        """Read field from FA file.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Raises:
            RuntimeError: validtime must be a datetime object

        Returns:
            np.darray: Field, surfex.Geo in read file

        """
        file_handler = Fa(self.filename)
        if validtime is None:
            pass
        elif isdatetime(validtime):
            raise RuntimeError("validtime must be a datetime object")

        field, geo_in = file_handler.field(var.varname, validtime)

        # Reshape to fortran 2D style
        logging.debug("field=%s, field.shape=%s", field, field.shape)
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest"):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geo): Surfex geometry for points.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type. Defaults to "nearest".

        Returns:
            tuple: Interpolated points, surfex.Interpolator

        """
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


class SurfFileTypeExtension(object):
    """Extension of a surfex file."""

    def __init__(self, csurf_filetype, masterodb=False, lfagmap=True):
        """Construct an extension object.

        Args:
            csurf_filetype (str): Surfex file type.
            masterodb (bool, optional): File created with masterodb. Defaults to False.
            lfagmap (bool, optional): File created with lfagmap=True. Defaults to True.

        """
        suffix = None
        extension = ""
        if csurf_filetype.lower() == "nc":
            extension = "nc"
        elif csurf_filetype.lower() == "ascii":
            extension = "txt"
        elif csurf_filetype.lower() == "fa":
            if lfagmap:
                extension = "sfx"
                if not masterodb:
                    suffix = ".fa"
            else:
                extension = "fa"

        if suffix is None:
            suffix = "." + extension

        self.type = extension
        self.suffix = suffix


class NetCDFSurfexFile(SurfexIO):
    """Reading surfex NetCDF time series output."""

    def __init__(self, filename, geo):
        """Construct a NetCDF surfex time series file.

        Args:
            filename (str): Name of file.
            geo (surfex.Geometry): Surfex geometry for time series.

        """
        self.file_handler = netCDF4.Dataset(filename, "r")
        SurfexIO.__init__(self, filename, geo, "nc")

    def read(self, var, times):
        """Read a field, return a 2D array.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            times (list): List of datetime.datetime objects

        Raises:
            ValueError: times must be list or tuple
            ValueError: patches must be list or tuple
            ValueError: patches must be list or tuple
            RuntimeError: Variable not found!
            NotImplementedError: Dimension not implemented
            RuntimeError: Mismatch in points

        Returns:
            (tuple): (np.array. surfex.Geometry)

        """
        layers = var.layers
        patches = var.patches

        if not isinstance(times, (list, tuple)):
            raise ValueError("times must be list or tuple")
        if not isinstance(layers, (list, tuple)):
            raise ValueError("patches must be list or tuple")
        if not isinstance(patches, (list, tuple)):
            raise ValueError("patches must be list or tuple")

        values = np.array([])
        times_read = []
        ndims = 0
        dim_indices = []
        mapping = {}
        npatch = 1

        if self.file_handler.variables[var.varname].shape[0] > 0:
            for dim in self.file_handler.variables[var.varname].dimensions:
                dimlen = self.file_handler.variables[var.varname].shape[ndims]
                this_dim = []

                if dim == "time":
                    mapping[0] = ndims
                    times_for_var = self.file_handler.variables["time"]
                    units = times_for_var.units
                    try:
                        t_cal = times_for_var.calendar
                    except AttributeError:  # Attribute doesn't exist
                        t_cal = "gregorian"  # or standard

                    indices = list(range(0, dimlen))
                    times_for_var = netCDF4.num2date(
                        times_for_var[indices], units=units, calendar=t_cal
                    )
                    if len(times) > 0:
                        for t_to_find, t_to_find_val in enumerate(times):
                            for tstep in range(0, len(indices)):
                                if times_for_var[tstep] == t_to_find_val:
                                    this_dim.append(tstep)
                                    times_read.append(times[t_to_find])
                    else:
                        times_read = times_for_var
                        this_dim = list(range(0, dimlen))

                elif dim == "Number_of_points":
                    mapping[1] = ndims
                    this_dim = list(range(0, dimlen))
                elif dim == "xx":
                    mapping[1] = ndims
                    this_dim = list(range(0, dimlen))
                elif dim == "yy":
                    mapping[2] = ndims
                    this_dim = list(range(0, dimlen))
                elif dim == "Number_of_Patches":
                    mapping[3] = ndims
                    npatch = dimlen
                    if len(patches) > 0:
                        npatch = len(patches)
                        this_dim = patches
                    else:
                        this_dim = list(range(0, dimlen))
                elif dim == "Number_of_Layers":
                    mapping[4] = ndims
                    npatch = dimlen
                    if len(layers) > 0:
                        this_dim = layers
                    else:
                        this_dim = list(range(0, dimlen))
                elif dim == "lon":
                    mapping[1] = ndims
                    this_dim = list(range(0, dimlen))
                elif dim == "lat":
                    mapping[2] = ndims
                    this_dim = list(range(0, dimlen))
                else:
                    raise NotImplementedError("Not implemented for: " + dim)

                dim_indices.append(this_dim)
                ndims = ndims + 1

            print(dim_indices)
            field = self.file_handler.variables[var.varname][dim_indices]

            # Add extra dimensions
            i = 0
            reverse_mapping = []
            for dim in range(0, 5):
                if dim not in mapping:
                    logging.debug("Adding dimension %s", dim)
                    field = np.expand_dims(field, len(dim_indices) + i)
                    reverse_mapping.append(len(dim_indices) + i)
                    i = i + 1
                else:
                    reverse_mapping.append(mapping[dim])

            # Transpose to 5D array
            field = np.transpose(field, reverse_mapping)
            npoints = self.geo.npoints * npatch

            # Create 2-D array with times and points as for the other formats
            for tstep in range(0, field.shape[0]):
                field2d = np.empty(npoints)
                i = 0
                for (patch,) in range(0, npatch):
                    if self.geo.mask is not None:
                        iii = 0
                        j = 0
                        # For some reason the NetCDF files have one less dimension in all
                        # dimensions than the PGD dimension and mask needs x first.
                        for xxx in range(-1, field.shape[1] + 1):
                            for yyy in range(-1, field.shape[2] + 1):
                                if xxx in range(0, field.shape[1]) and yyy in range(
                                    0, field.shape[2]
                                ):
                                    if self.geo.mask[j] == iii:
                                        field2d[i] = np.nan
                                        if field[tstep, xxx, yyy, patch] != np.nan:
                                            field2d[i] = field[tstep, xxx, yyy, patch]
                                        i = i + 1
                                        j = j + 1
                                iii = iii + 1
                    else:
                        for yyy in range(0, field.shape[2]):
                            for xxx in range(0, field.shape[1]):
                                field2d[i] = np.nan
                                if field[tstep, xxx, yyy, patch] != np.nan:
                                    field2d[i] = field[tstep, xxx, yyy, patch]
                                i = i + 1
                if i != npoints:
                    raise RuntimeError(
                        "Mismatch in points " + str(i) + "!=" + str(npoints)
                    )

                values = np.append(values, field2d)
            # Re-shape to proper format
            values = np.reshape(values, [field.shape[0], npoints])

        else:
            raise RuntimeError("Variable " + var.varname + " not found!")

        return values, self.geo

    def field(self, var, validtime=None):
        """Read field.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Returns:
            tuple: (np.array, surfex.Geometry)

        """
        if validtime is None:
            validtime = []
        else:
            validtime = [validtime]

        field, geo_in = self.read(var, validtime)
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest"):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geometry): Surfex geometry to interpolate to.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type.. Defaults to "nearest".

        Returns:
            tuple: points, surfex.Interpolator

        """
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


class TexteSurfexFile(SurfexIO):
    """Reading surfex TEXTE output."""

    def __init__(self, filename, geo):
        """Construct the Texte file.

        Args:
            filename (str): Filename
            geo (surfex.Geometry): Geometry

        """
        self.file = None
        SurfexIO.__init__(self, filename, geo, "TXT")

    def read(self, variable, times):
        """Read file.

        Args:
            variable (SurfexFileVariable): Variable in surfex file.
            times (list): List of datetime.datetime to read.

        Raises:
            RuntimeError: Basetime must be set for TEXTE
            RuntimeError: Interval must be set for TEXTE
            RuntimeError: times must be list or tuple
            RuntimeError: Dimension of domain does not match end of line

        Returns:
            tuple: (np.array, surfex.Geometry)

        """
        self.file = open(self.filename, mode="r", encoding="utf-8")

        base_time = variable.basetime
        interval = variable.interval
        npatch = variable.patches

        if base_time is None:
            raise RuntimeError("Basetime must be set for TEXTE")
        if interval is None:
            raise RuntimeError("Interval must be set for TEXTE")

        if not isinstance(times, (list, tuple)):
            raise RuntimeError("times must be list or tuple")

        values = np.array([])
        times_read = np.array([])
        end_of_line = self.geo.npoints * npatch
        this_time = np.empty(self.geo.npoints * npatch)

        tstep = 0
        col = 0
        for line in self.file.read().splitlines():

            words = line.split()
            if len(words) > 0:
                for i, word in enumerate(words):
                    val = float(word.replace("D", "E"))
                    if val == 1e20:
                        val = np.nan
                    this_time[col] = val

                    col = col + 1
                    if col == end_of_line:
                        if (
                            times is None
                            or (base_time + as_timedelta(seconds=(tstep * interval)))
                            in times
                        ):
                            values = np.append(values, this_time)
                            times_read = np.append(
                                times_read,
                                base_time + as_timedelta(seconds=(tstep * interval)),
                            )

                        tstep = tstep + 1
                        col = 0
                        this_time[:] = np.nan
                        if i != len(words) - 1:
                            raise RuntimeError(
                                "Dimension of domain does not match end of line! "
                                + str(i)
                                + " != "
                                + str(len(words) - 1)
                            )

        if times_read.shape[0] > 0:
            values = np.asarray(values)
            values = np.reshape(values, [times_read.shape[0], this_time.shape[0]])
        else:
            logging.info("No data found!")

        self.file.close()
        return values, self.geo

    def field(self, var, validtime=None):
        """Read field.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Returns:
            tuple: (np.array, surfex.Geometry)

        """
        if validtime is not None:
            validtime = [validtime]

        field, geo_in = self.read(var, validtime)
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation="nearest"):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geometry): Geometry in file.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type. Defaults to "nearest".

        Returns:
            tuple: (np.array, surfex.Interpolator)

        """
        field, geo_in = self.field(var, validtime=validtime)
        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


class ForcingFileNetCDF(SurfexIO):
    """Forcing netCDF file."""

    def __init__(self, fname, geo):
        """Construct forcing netcdf file.

        Args:
            fname (str): File name.
            geo (surfex.Geo): Surfex input geometry.

        """
        self.fname = fname
        self.file_handler = netCDF4.Dataset(fname, "r")
        self.lons = self.file_handler.variables["LON"]
        self.lats = self.file_handler.variables["LAT"]
        SurfexIO.__init__(self, fname, geo, "nc")

    def read_field(self, variable, times):
        """Read file.

        Args:
            variable (_type_): _description_
            times (_type_): _description_

        Raises:
            RuntimeError: You must set time!
            RuntimeError: No points found
            RuntimeError: Valid time not found in file!

        Returns:
            tuple: field, geo

        """
        var = variable.varname
        field = None
        if var in self.file_handler.variables:
            if self.file_handler.variables[var].shape[0] > 0:
                if len(self.file_handler.variables[var].dimensions) == 1:
                    dimlen = self.file_handler.variables[var].shape[0]
                    field = self.file_handler.variables[var][0:dimlen]
                else:
                    if len(times) == 0:
                        raise RuntimeError("You must set time!")

                    times_read = []
                    ndims = 0
                    npoints = 0
                    for dim in self.file_handler.variables[var].dimensions:
                        dimlen = self.file_handler.variables[var].shape[ndims]

                        if dim == "time":
                            times_for_var = self.file_handler.variables["time"]
                            units = times_for_var.units
                            try:
                                t_cal = times_for_var.calendar
                            except AttributeError:  # Attribute doesn't exist
                                t_cal = "gregorian"  # or standard

                            indices = list(range(0, dimlen))
                            times_for_var = netCDF4.num2date(
                                times_for_var[indices], units=units, calendar=t_cal
                            )
                            for times_to_read_val in times:
                                for tstep, times_for_var_val in enumerate(times_for_var):
                                    test_time = times_for_var_val.strftime("%Y%m%d%H")
                                    test_time = as_datetime(test_time)
                                    if test_time == times_to_read_val:
                                        times_read.append(tstep)
                                        logging.debug("%s %s", tstep, times_to_read_val)
                        else:
                            npoints = dimlen

                        ndims = ndims + 1

                    if npoints == 0:
                        raise RuntimeError("No points found")

                    if len(times_read) == 0 and len(times) > 0:
                        logging.error("%s", times)
                        raise RuntimeError("Valid time not found in file!")

                    field = self.file_handler.variables[var][times_read, 0:npoints]
        else:
            logging.warning("Variable %s not found!", var)
        return field, self.geo

    def field(self, var, validtime=None):
        """Read field.

        Args:
            var (_type_): _description_
            validtime (list, optional): Validtime. Defaults to None.

        Returns:
            tuple: field, geo

        """
        if validtime is None:
            validtime = []
        else:
            validtime = [validtime]

        field, geo_in = self.read_field(var, validtime)
        # Reshape to fortran 2D style
        if field is not None:
            field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo_out, validtime=None, interpolation=None):
        """Read points.

        Args:
            var (SurfexFileVariable): Variable in surfex file.
            geo_out (surfex.Geo): Surfex geometry for points.
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation type. Defaults to "nearest".

        Returns:
            tuple: Interpolated points, surfex.Interpolator

        """
        field, geo_in = self.field(var, validtime=validtime)

        points, interpolator = SurfexIO.interpolate_field(
            field, geo_in, geo_out, interpolation=interpolation
        )
        return points, interpolator


def read_surfex_field(
    varname,
    filename,
    validtime=None,
    basetime=None,
    patches=-1,
    layers=-1,
    fileformat=None,
    filetype=None,
    geo=None,
    datatype=None,
    interval=None,
    tiletype="FULL",
):
    """Read surfex field.

    Args:
        varname (_type_): _description_
        filename (_type_): _description_
        validtime (_type_, optional): _description_. Defaults to None.
        basetime (_type_, optional): _description_. Defaults to None.
        patches (int, optional): _description_. Defaults to -1.
        layers (int, optional): _description_. Defaults to -1.
        fileformat (_type_, optional): _description_. Defaults to None.
        filetype (_type_, optional): _description_. Defaults to None.
        geo (_type_, optional): _description_. Defaults to None.
        datatype (_type_, optional): _description_. Defaults to None.
        interval (_type_, optional): _description_. Defaults to None.
        tiletype(str, optional): Tiletype. Defaults to "FULL".

    Raises:
        RuntimeError: Not implemented and geo is None
        RuntimeError: You need to provide a geo object.

    Returns:
        field (np.ndarray): Field

    """
    if fileformat is None:
        fileformat, filetype = guess_file_format(filename, filetype)

    if filetype == "surf":
        if fileformat.lower() == "ascii":
            geo = AsciiSurfexFile(filename).geo
        elif fileformat.lower() == "nc":
            geo = NCSurfexFile(filename).geo
        else:
            if geo is None:
                raise RuntimeError("Not implemented and geo is None")
    elif geo is None:
        raise RuntimeError(
            "You need to provide a geo object. Filetype is: " + str(filetype)
        )

    sfx_io = get_surfex_io_object(
        filename, filetype=filetype, fileformat=fileformat, geo=geo
    )
    var = SurfexFileVariable(
        varname,
        validtime=validtime,
        patches=patches,
        layers=layers,
        basetime=basetime,
        interval=interval,
        datatype=datatype,
        tiletype=tiletype,
    )
    field, __ = sfx_io.field(var, validtime=validtime)
    return field


def read_surfex_points(
    varname,
    filename,
    geo_out,
    validtime=None,
    basetime=None,
    patches=-1,
    layers=-1,
    fileformat=None,
    filetype=None,
    geo=None,
    datatype=None,
    interval=None,
    interpolation="nearest",
    tiletype="FULL",
):
    """Read surfex points.

    Args:
        varname (str): Variable name.
        filename (str): File name.
        geo_out (surfex.Geo): Surfex geometry
        validtime (datetime.datetime, optional): Valid time. Defaults to None.
        basetime (datetime.datetime, optional): Base time. Defaults to None.
        patches (int, optional): Number of patches. Defaults to -1.
        layers (int, optional): Numbers of layers. Defaults to -1.
        fileformat (str, optional): File format. Defaults to None.
        filetype (str, optional): File type. Defaults to None.
        geo (surfex.Geo, optional): Input geometry if needed. Defaults to None.
        datatype (str, optional): Data type. Defaults to None.
        interval (int, optional): Interval between times. Defaults to None.
        interpolation (str, optional): Interpolation method. Defaults to "nearest".
        tiletype(str, optional): Tiletype. Defaults to "FULL".

    Raises:
        NotImplementedError: _description_
        RuntimeError: _description_

    Returns:
        np.darray: Field

    """
    if fileformat is None:
        fileformat, filetype = guess_file_format(filename, filetype)

    if filetype == "surf":
        if fileformat.lower() == "ascii":
            geo = AsciiSurfexFile(filename).geo
        elif fileformat.lower() == "nc":
            geo = NCSurfexFile(filename).geo
        else:
            if geo is None:
                raise NotImplementedError(
                    f"{fileformat} is not implemented and geo is None"
                )
    elif geo is None:
        raise RuntimeError(
            "You need to provide a geo object. Filetype is: " + str(filetype)
        )

    sfx_io = get_surfex_io_object(
        filename, filetype=filetype, fileformat=fileformat, geo=geo
    )
    var = SurfexFileVariable(
        varname,
        validtime=validtime,
        patches=patches,
        layers=layers,
        basetime=basetime,
        interval=interval,
        datatype=datatype,
        tiletype=tiletype,
    )
    field, geo_out = sfx_io.points(
        var, geo_out, validtime=validtime, interpolation=interpolation
    )
    return field



#!/usr/bin/env python3
"""Implement helper routines to deal with dates and times."""
from datetime import date, datetime, timedelta


# TODO use ISO times
def as_datetime(dtg):
    """Convert string to datetime."""
    if len(dtg) == 10:
        fmt = "%Y%m%d%H"
    elif len(dtg) == 12:
        fmt = "%Y%m%d%H%M"
    elif len(dtg) == 14:
        fmt = "%Y%m%d%H%M%S"
    else:
        raise RuntimeError(f"dtg={dtg} len(dtg) is {len(dtg)}")

    return datetime.strptime(dtg, fmt)


def as_datetime_string(dtg):
    """Convert string to datetime."""
    fmt = "%Y%m%d%H%M%S"
    return dtg.strftime(fmt)


def as_timedelta(seconds=0):
    """Convert seconds to timedelta."""
    return timedelta(seconds=seconds)


def fromtimestamp(validtime):
    """Convert timestamp to validtime."""
    return datetime.fromtimestamp(validtime)


def utcfromtimestamp(epochtime):
    """Convert timestamp to validtime."""
    return datetime.utcfromtimestamp(epochtime)


def isdatetime(obj):
    """Check if is a datetime objects."""
    return isinstance(obj, date)


def as_datetime_args(year=None, month=None, day=None, hour=0, minute=0, second=0):
    """Set datetime object from args."""
    return datetime(
        year=year, month=month, day=day, hour=hour, minute=minute, second=second
    )



"""Misc."""
import collections
import logging
import os

import toml


def merge_toml_env(old_env, mods):
    """Merge."""
    return deep_update(old_env, mods)


def merge_toml_env_from_files(toml_files):
    """Merge."""
    merged_env = {}
    for toml_file in toml_files:
        if os.path.exists(toml_file):
            with open(toml_file, mode="r", encoding="utf-8") as file_handler:
                modification = toml.load(file_handler)
            merged_env = merge_toml_env(merged_env, modification)
        else:
            print("WARNING: File not found " + toml_file)
    return merged_env


def deep_update(source, overrides):
    """Update a nested dictionary or similar mapping.

    Modify ``source`` in place.

    Args:
        source(dict): Source data
        overrides(dict): Delta data to override

    Returns:
        source(dict): Updated dict
    """
    for key, value in overrides.items():
        if isinstance(value, collections.abc.Mapping) and value:
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            override = overrides[key]
            source[key] = override
    return source


def remove_existing_file(f_in, f_out):
    """Remove existing file.

    Args:
        f_in (_type_): _description_
        f_out (_type_): _description_

    Raises:
        FileNotFoundError: _description_
        IsADirectoryError: _description_

    """
    if f_in is None:
        raise FileNotFoundError("Input file not set")
    # If files are not the same file
    if os.path.abspath(f_in) != os.path.abspath(f_out):
        if os.path.isdir(f_out):
            raise IsADirectoryError(
                f_out + " is a directory! Please remove it if desired"
            )
        if os.path.islink(f_out):
            os.unlink(f_out)
        if os.path.isfile(f_out):
            os.remove(f_out)
    # files have the same path. Remove if it is a symlink
    else:
        if os.path.islink(f_out):
            os.unlink(f_out)


def parse_filepattern(file_pattern, basetime, validtime):
    """Parse the file pattern.

    Args:
        file_pattern (str): File pattern.
        basetime (datetime.datetime): Base time.
        validtime (datetime.datetime): Valid time.

    Returns:
        str: File name

    """
    if basetime is None or validtime is None:
        return file_pattern

    logging.debug(
        "file_pattern=%s basetime=%s validtime=%s", file_pattern, basetime, validtime
    )
    file_name = str(file_pattern)
    year = basetime.strftime("%Y")
    year2 = basetime.strftime("%y")
    month = basetime.strftime("%m")
    day = basetime.strftime("%d")
    hour = basetime.strftime("%H")
    mins = basetime.strftime("%M")
    d_t = validtime - basetime
    ll_d = f"{int(d_t.seconds / 3600):d}"
    ll_2 = f"{int(d_t.seconds / 3600):02d}"
    ll_3 = f"{int(d_t.seconds / 3600):03d}"
    ll_4 = f"{int(d_t.seconds / 3600):04d}"
    file_name = file_name.replace("@YYYY@", year)
    file_name = file_name.replace("@YY@", year2)
    file_name = file_name.replace("@MM@", month)
    file_name = file_name.replace("@DD@", day)
    file_name = file_name.replace("@HH@", hour)
    file_name = file_name.replace("@mm@", mins)
    file_name = file_name.replace("@L@", ll_d)
    file_name = file_name.replace("@LL@", ll_2)
    file_name = file_name.replace("@LLL@", ll_3)
    file_name = file_name.replace("@LLLL@", ll_4)
    return file_name



"""Ecoclimap handling."""
import logging
import os


class ExternalSurfexInputFile(object):
    """Wrapper around external input data to surfex.

    Can have special treatment for each format.
    Uses internally the SystemFilePaths class
    """

    def __init__(self, system_file_paths):
        """Construct ExternalSurfexInputFile.

        Args:
            system_file_paths (surfex.SystemFilePaths): Match system specific files.

        """
        self.system_file_paths = system_file_paths

    def set_input_data_from_format(
        self,
        dtype,
        fname,
        default_dir=None,
        check_existence=False,
        check_parsing=True,
        validtime=None,
        basedtg=None,
        mbr=None,
        tstep=None,
        pert=None,
        var=None,
        system_variables=None,
        linkbasename=None,
    ):
        """Set input data based on format.

        Args:
            dtype (_type_): _description_
            fname (_type_): _description_
            default_dir (_type_, optional): _description_. Defaults to None.
            check_existence (bool, optional): _description_. Defaults to False.
            check_parsing (bool, optional): _description_. Defaults to True.
            validtime (_type_, optional): _description_. Defaults to None.
            basedtg (_type_, optional): _description_. Defaults to None.
            mbr (_type_, optional): _description_. Defaults to None.
            tstep (_type_, optional): _description_. Defaults to None.
            pert (_type_, optional): _description_. Defaults to None.
            var (_type_, optional): _description_. Defaults to None.
            system_variables (_type_, optional): _description_. Defaults to None.
            linkbasename (_type_, optional): _description_. Defaults to None.

        Returns:
            dict: File name mappings.

        """
        fname_with_path = self.system_file_paths.get_system_file(
            dtype,
            fname,
            default_dir=default_dir,
            check_existence=check_existence,
            check_parsing=check_parsing,
            validtime=validtime,
            basedtg=basedtg,
            mbr=mbr,
            tstep=tstep,
            pert=pert,
            var=var,
            system_variables=system_variables,
        )

        if fname.endswith(".dir"):
            basename = os.path.splitext(os.path.basename(fname))[0]

            basedir = self.system_file_paths.get_system_path(
                dtype,
                default_dir=default_dir,
                check_existence=check_existence,
                check_parsing=check_parsing,
                validtime=validtime,
                basedtg=basedtg,
                mbr=mbr,
                tstep=tstep,
                pert=pert,
                var=var,
            )
            logging.debug("%s %s %s", basename, basedir, fname_with_path)
            hdr_file = basedir + "/" + basename + ".hdr"
            dir_file = basedir + "/" + basename + ".dir"
            if linkbasename is None:
                linkbasename = basename
            return {linkbasename + ".hdr": hdr_file, linkbasename + ".dir": dir_file}
        elif fname.endswith(".json"):
            return {}
        else:
            return {fname: fname_with_path}


class Ecoclimap(object):
    """Ecoclimap."""

    def __init__(self, config, system_file_paths=None):
        """Construct ecoclimap data object.

        Args:
            config (surfex.Configuration): Surfex configuration.
            system_file_paths (surfex.SystemFilePaths, optional): Mapping of local file structure
                                                                  to look for inut files.
                                                                  Defaults to None.

        """
        self.config = config
        self.system_file_paths = system_file_paths
        self.cover_dir = "ecoclimap_cover_dir"
        self.bin_dir = "ecoclimap_bin_dir"
        self.ecoclimap_files = [
            "ecoclimapI_covers_param.bin",
            "ecoclimapII_af_covers_param.bin",
            "ecoclimapII_eu_covers_param.bin",
        ]
        self.decadal_data_types = None

    def set_input(self, check_existence=True):
        """Set input.

        Args:
            check_existence (bool, optional): _description_. Defaults to True.

        Raises:
            RuntimeError: System file path must be set for this method

        Returns:
            dict: File mappings.

        """
        if self.system_file_paths is None:
            raise RuntimeError("System file path must be set for this method")

        data = {}
        for fname in self.ecoclimap_files:
            fname_data = self.system_file_paths.get_system_file(
                self.bin_dir,
                fname,
                default_dir="climdir",
                check_existence=check_existence,
            )
            data.update({fname: fname_data})
        return data

    def set_bin_files(self, check_existence=True):
        """Set bin files.

        Args:
            check_existence (bool, optional): Check if files exist. Defaults to True.

        Returns:
            dict: File mappings.

        """
        return self.set_input(check_existence=check_existence)


class EcoclimapSG(Ecoclimap):
    """Ecoclimap SG."""

    def __init__(self, config, system_file_paths=None, veg_types=20, decades=36):
        """Construct ecoclimap SG.

        Args:
            config (_type_): _description_
            system_file_paths (_type_, optional): _description_. Defaults to None.
            veg_types (int, optional): _description_. Defaults to 20.
            decades (int, optional): _description_. Defaults to 36.

        """
        Ecoclimap.__init__(self, config, system_file_paths=system_file_paths)
        self.veg_types = veg_types
        self.decades = decades
        self.cover_file = self.config.get_setting("SURFEX#COVER#SG")
        self.cover_dir = "ecoclimap_sg_cover_dir"
        self.decadal_data_types = [
            "ALBNIR_SOIL",
            "ALBNIR_VEG",
            "ALBVIS_SOIL",
            "ALBVIS_VEG",
            "LAI",
        ]

    def set_bin_files(self, check_existence=True):
        """set_bin_files not used for SG."""

    def set_input(self, check_existence=True):
        """Set input data.

        Args:
            check_existence (bool, optional): Check if files are existing. Defaults to True.

        Raises:
            RuntimeError: System file path must be set for this method

        Returns:
            dict: Mapping of files.

        """
        if self.system_file_paths is None:
            raise RuntimeError("System file path must be set for this method")

        data = {}
        tree_height_dir = "tree_height_dir"
        fname = self.config.get_setting("SURFEX#COVER#H_TREE")
        if fname != "" and fname is not None:
            ext_data = ExternalSurfexInputFile(self.system_file_paths)
            data.update(
                ext_data.set_input_data_from_format(
                    tree_height_dir, fname, check_existence=check_existence
                )
            )

        decadal_data_types = [
            "ALBNIR_SOIL",
            "ALBNIR_VEG",
            "ALBVIS_SOIL",
            "ALBVIS_VEG",
            "LAI",
        ]
        for decadal_data_type in decadal_data_types:
            for __ in range(1, self.veg_types + 1):
                for decade in range(1, self.decades + 1):
                    filepattern = self.config.get_setting(
                        "SURFEX#COVER#" + decadal_data_type, check_parsing=False
                    )
                    fname = self.parse_fnames(filepattern, decade)
                    dtype = decadal_data_type.lower() + "_dir"
                    ext_data = ExternalSurfexInputFile(self.system_file_paths)
                    dat = ext_data.set_input_data_from_format(
                        dtype, fname, check_existence=check_existence
                    )
                    data.update(dat)
        return data

    @staticmethod
    def parse_fnames(filepattern, decade):
        """Parse file names."""
        filename = filepattern
        decade = decade - 1
        mmm = int(decade / 3) + 1
        cmm = f"{mmm:02d}"
        cdd = ((decade % 3) * 10) + 5
        cdd = f"{cdd:02d}"
        filename = filename.replace("@MM@", str(cmm))
        filename = filename.replace("@CDD@", str(cdd))
        return filename
